from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals


class ExpertsConfig(AppConfig):
    name = 'waldur_mastermind.experts'
    verbose_name = 'Experts'

    def ready(self):
        from waldur_core.structure import models as structure_models
        from waldur_mastermind.invoices import registrators as invoices_registrators
        from waldur_mastermind.support.models import Comment

        from . import handlers, registrators, quotas

        ExpertRequest = self.get_model('ExpertRequest')
        ExpertBid = self.get_model('ExpertBid')
        ExpertContract = self.get_model('ExpertContract')

        invoices_registrators.RegistrationManager.add_registrator(
            ExpertRequest,
            registrators.ExpertRequestRegistrator
        )

        quotas.add_quota_field()

        signals.post_save.connect(
            handlers.update_project_quota_when_request_is_saved,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_project_quota_when_request_is_saved',
        )

        signals.pre_delete.connect(
            handlers.update_project_quota_when_request_is_deleted,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_project_quota_when_request_is_deleted',
        )

        signals.post_save.connect(
            handlers.update_customer_quota_when_request_is_saved,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_customer_quota_when_request_is_saved',
        )

        signals.pre_delete.connect(
            handlers.update_customer_quota_when_request_is_deleted,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_customer_quota_when_request_is_deleted',
        )

        signals.post_save.connect(
            handlers.add_completed_expert_request_to_invoice,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'add_completed_expert_request_to_invoice',
        )

        signals.pre_delete.connect(
            handlers.terminate_invoice_when_expert_request_deleted,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'terminate_invoice_when_expert_request_deleted',
        )

        signals.post_save.connect(
            handlers.log_expert_request_creation,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'log_expert_request_creation',
        )

        signals.post_save.connect(
            handlers.log_expert_request_state_changed,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'log_expert_request_state_changed',
        )

        signals.post_save.connect(
            handlers.log_expert_bid_creation,
            sender=ExpertBid,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'log_expert_bid_creation',
        )

        signals.post_save.connect(
            handlers.notify_expert_providers_about_new_request,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'notify_expert_providers_about_new_request',
        )

        signals.post_save.connect(
            handlers.notify_customer_owners_about_new_bid,
            sender=ExpertBid,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'notify_customer_owners_about_new_bid',
        )

        signals.post_save.connect(
            handlers.notify_customer_owners_about_new_contract,
            sender=ExpertContract,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'notify_customer_owners_about_new_contract',
        )

        signals.post_save.connect(
            handlers.update_customer_quota_when_contract_is_created,
            sender=ExpertContract,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_customer_quota_when_contract_is_created',
        )

        signals.pre_delete.connect(
            handlers.update_customer_quota_when_contract_is_deleted,
            sender=ExpertContract,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_customer_quota_when_contract_is_deleted',
        )

        signals.post_save.connect(
            handlers.set_project_name_on_expert_request_creation,
            sender=ExpertRequest,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'set_project_name_on_expert_request_creation',
        )

        signals.post_save.connect(
            handlers.update_expert_request_on_project_name_update,
            sender=structure_models.Project,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_expert_request_on_project_name_update',
        )

        signals.post_save.connect(
            handlers.set_team_name_on_expert_contract_creation,
            sender=ExpertContract,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'set_team_name_on_expert_contract_creation',
        )

        signals.post_save.connect(
            handlers.update_expert_contract_on_project_name_update,
            sender=structure_models.Project,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'update_expert_contract_on_project_name_update',
        )

        signals.post_save.connect(
            handlers.send_expert_comment_added_notification,
            sender=Comment,
            dispatch_uid='waldur_mastermind.experts.handlers.'
                         'send_expert_comment_added_notification',
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0052_customer_subnets'),
        ('openstack_tenant', '0028_remove_duplicate_security_groups_networks_subnets'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='network',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='securitygroup',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='subnet',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



from __future__ import unicode_literals

import logging

from django.db import transaction
from django.utils import timezone
import requests
import six

from waldur_core.structure import log_backend_action, ServiceBackend, ServiceBackendError
from waldur_core.structure.utils import (
    update_pulled_fields, handle_resource_not_found, handle_resource_update_success)

from . import models
from .client import RijkscloudClient


logger = logging.getLogger(__name__)


class RijkscloudBackendError(ServiceBackendError):
    pass


class RijkscloudBackend(ServiceBackend):

    def __init__(self, settings):
        """
        :type settings: :class:`waldur_core.structure.models.ServiceSettings`
        """
        self.settings = settings
        self.client = RijkscloudClient(
            userid=settings.username,
            apikey=settings.token,
        )

    def ping(self, raise_exception=False):
        try:
            self.client.list_flavors()
        except requests.RequestException as e:
            if raise_exception:
                six.reraise(RijkscloudBackendError, e)
            return False
        else:
            return True

    def sync(self):
        self.pull_flavors()
        self.pull_floating_ips()
        self.pull_networks()
        self.pull_volumes()
        self.pull_instances()

    def _get_current_properties(self, model):
        return {p.backend_id: p for p in model.objects.filter(settings=self.settings)}

    def _get_backend_resource(self, model, resources):
        registered_backend_ids = model.objects.filter(
            service_project_link__service__settings=self.settings).values_list('backend_id', flat=True)
        return [instance for instance in resources if instance.backend_id not in registered_backend_ids]

    def pull_flavors(self):
        try:
            flavors = self.client.list_flavors()
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

        with transaction.atomic():
            cur_flavors = self._get_current_properties(models.Flavor)
            for backend_flavor in flavors:
                cur_flavors.pop(backend_flavor['name'], None)
                models.Flavor.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_flavor['name'],
                    defaults={
                        'name': backend_flavor['name'],
                        'cores': backend_flavor['vcpus'],
                        'ram': backend_flavor['ram'],
                    })

            models.Flavor.objects.filter(backend_id__in=cur_flavors.keys()).delete()

    def pull_volumes(self):
        backend_volumes = self.get_volumes()
        volumes = models.Volume.objects.filter(
            service_project_link__service__settings=self.settings,
            state__in=[models.Volume.States.OK, models.Volume.States.ERRED]
        )
        backend_volumes_map = {backend_volume.backend_id: backend_volume for backend_volume in backend_volumes}
        for volume in volumes:
            try:
                backend_volume = backend_volumes_map[volume.backend_id]
            except KeyError:
                handle_resource_not_found(volume)
            else:
                update_pulled_fields(volume, backend_volume, models.Volume.get_backend_fields())
                handle_resource_update_success(volume)

    def get_volumes(self):
        try:
            backend_volumes = self.client.list_volumes()
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)
        else:
            return [self._backend_volume_to_volume(backend_volume)
                    for backend_volume in backend_volumes]

    def _backend_volume_to_volume(self, backend_volume):
        return models.Volume(
            name=backend_volume['name'],
            backend_id=backend_volume['name'],
            size=backend_volume['size'] * 1024,
            metadata=backend_volume['metadata'],
            runtime_state=backend_volume['status'],
            state=models.Volume.States.OK,
        )

    @log_backend_action()
    def pull_volume(self, volume, update_fields=None):
        import_time = timezone.now()
        imported_volume = self.import_volume(volume.backend_id, save=False)

        volume.refresh_from_db()
        if volume.modified < import_time:
            if not update_fields:
                update_fields = models.Volume.get_backend_fields()

            update_pulled_fields(volume, imported_volume, update_fields)

    def import_volume(self, backend_volume_id, save=True, service_project_link=None):
        try:
            backend_volume = self.client.get_volume(backend_volume_id)
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)
        volume = self._backend_volume_to_volume(backend_volume)
        if service_project_link is not None:
            volume.service_project_link = service_project_link
        if save:
            volume.save()

        return volume

    def get_volumes_for_import(self):
        return self._get_backend_resource(models.Volume, self.get_volumes())

    @log_backend_action()
    def pull_volume_runtime_state(self, volume):
        try:
            backend_volume = self.client.get_volume(volume.backend_id)
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)
        else:
            if backend_volume['status'] != volume.runtime_state:
                volume.runtime_state = backend_volume['status']
                volume.save(update_fields=['runtime_state'])

    @log_backend_action()
    def delete_volume(self, volume):
        try:
            self.client.delete_volume(volume.backend_id)
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

    @log_backend_action('check is volume deleted')
    def is_volume_deleted(self, instance):
        try:
            self.client.get_volume(instance.backend_id)
            return False
        except requests.RequestException as e:
            if e.response.status_code == 404:
                return True
            else:
                six.reraise(RijkscloudBackendError, e)

    def pull_instances(self):
        backend_instances = self.get_instances()
        instances = models.Instance.objects.filter(
            service_project_link__service__settings=self.settings,
            state__in=[models.Instance.States.OK, models.Instance.States.ERRED],
        )
        backend_instances_map = {backend_instance.backend_id: backend_instance
                                 for backend_instance in backend_instances}
        for instance in instances:
            try:
                backend_instance = backend_instances_map[instance.backend_id]
            except KeyError:
                handle_resource_not_found(instance)
            else:
                self.update_instance_fields(instance, backend_instance)
                handle_resource_update_success(instance)

    def update_instance_fields(self, instance, backend_instance):
        # Preserve flavor fields in Waldur database if flavor is deleted in Rijkscloud
        fields = set(models.Instance.get_backend_fields())
        flavor_fields = {'flavor_name', 'ram', 'cores'}
        if not backend_instance.flavor_name:
            fields = fields - flavor_fields
        fields = list(fields)

        update_pulled_fields(instance, backend_instance, fields)

    def get_instances(self):
        try:
            backend_instances = self.client.list_instances()
            backend_flavors = self.client.list_flavors()
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

        backend_flavors_map = {flavor['name']: flavor for flavor in backend_flavors}
        instances = []
        for backend_instance in backend_instances:
            instance_flavor = backend_flavors_map.get(backend_instance['flavor'])
            instances.append(self._backend_instance_to_instance(backend_instance, instance_flavor))
        return instances

    def _backend_instance_to_instance(self, backend_instance, backend_flavor=None):
        instance = models.Instance(
            name=backend_instance['name'],
            state=models.Instance.States.OK,
            backend_id=backend_instance['name'],
        )
        if backend_flavor:
            instance.flavor_name = backend_flavor['name']
            instance.cores = backend_flavor['vcpus']
            instance.ram = backend_flavor['ram']

        # It is assumed that addresses list contains either one or two items, where
        # first item is internal IP address and second item is floating IP address.
        # This code does not handle case when internal subnet CIDR overlaps with other internal subnet.
        addresses = backend_instance['addresses']
        instance.internal_ip = models.InternalIP.objects.filter(
            settings=self.settings, address=addresses[0]).first()
        if len(addresses) == 2:
            instance.floating_ip = models.FloatingIP.objects.filter(
                settings=self.settings, address=addresses[1]).first()
        return instance

    @log_backend_action()
    def pull_instance(self, instance, update_fields=None):
        import_time = timezone.now()
        imported_instance = self.import_instance(instance.backend_id, save=False)

        instance.refresh_from_db()
        if instance.modified < import_time:
            if update_fields is None:
                update_fields = models.Instance.get_backend_fields()
            update_pulled_fields(instance, imported_instance, update_fields)

    def import_instance(self, backend_instance_id, save=True, service_project_link=None):
        try:
            backend_instance = self.client.get_instance(backend_instance_id)
            flavor = self.client.get_flavor(backend_instance['flavor'])
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

        instance = self._backend_instance_to_instance(backend_instance, flavor)
        if service_project_link:
            instance.service_project_link = service_project_link
        if save:
            instance.save()
        return instance

    def get_instances_for_import(self):
        return self._get_backend_resource(models.Instance, self.get_instances())

    @log_backend_action()
    def create_volume(self, volume):
        kwargs = {
            'size': max(1, int(volume.size / 1024)),
            'name': volume.name,
            'description': volume.description,
        }
        try:
            self.client.create_volume(kwargs)
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

        backend_volume = self.client.get_volume(volume.name)
        volume.backend_id = volume.name
        volume.runtime_state = backend_volume['status']
        volume.save()
        return volume

    @log_backend_action()
    def create_instance(self, instance):
        # It's impossible to specify custom security group
        # because Rijkscloud API does not provide security groups API yet.
        kwargs = {
            'name': instance.name,
            'flavor': instance.flavor_name,
            'userdata': instance.user_data or 'normal',
            'interfaces': [
                {
                    'subnets': [
                        {
                            'ip': instance.internal_ip.address,
                            'name': instance.internal_ip.subnet.name,
                        }
                    ],
                    'network': instance.internal_ip.subnet.network.name,
                    'security_groups': ['any-any']
                }
            ]
        }

        if instance.floating_ip:
            kwargs['interfaces'][0]['float'] = instance.floating_ip.address

        try:
            self.client.create_instance(kwargs)
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

        instance.backend_id = instance.name
        instance.save()
        return instance

    @log_backend_action()
    def delete_instance(self, instance):
        try:
            self.client.delete_instance(instance.backend_id)
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)

    @log_backend_action('check is instance deleted')
    def is_instance_deleted(self, instance):
        try:
            self.client.get_instance(instance.backend_id)
            return False
        except requests.RequestException as e:
            if e.response.status_code == 404:
                return True
            else:
                six.reraise(RijkscloudBackendError, e)

    def pull_floating_ips(self):
        try:
            backend_floating_ips = self.client.list_floatingips()
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)
            return

        with transaction.atomic():
            cur_floating_ips = self._get_current_properties(models.FloatingIP)
            for backend_fip in backend_floating_ips:
                cur_floating_ips.pop(backend_fip['float_ip'], None)
                models.FloatingIP.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_fip['float_ip'],
                    defaults={
                        'address': backend_fip['float_ip'],
                        'is_available': backend_fip['available'],
                    })

            models.FloatingIP.objects.filter(backend_id__in=cur_floating_ips.keys()).delete()

    def pull_networks(self):
        try:
            backend_networks = self.client.list_networks()
        except requests.RequestException as e:
            six.reraise(RijkscloudBackendError, e)
            return

        with transaction.atomic():
            current_networks = self._get_current_properties(models.Network)
            for backend_network in backend_networks:
                current_networks.pop(backend_network['name'], None)
                network, _ = models.Network.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_network['name'],
                    defaults=dict(name=backend_network['name']),
                )
                self.pull_subnets(network, backend_network['subnets'])

            models.Network.objects.filter(backend_id__in=current_networks.keys()).delete()

    def pull_subnets(self, network, backend_subnets):
        for backend_subnet in backend_subnets:
            gateway_ip = backend_subnet['gateway_ip']
            if isinstance(gateway_ip, list):
                gateway_ip = gateway_ip[0]
            subnet, _ = models.SubNet.objects.update_or_create(
                settings=self.settings,
                network=network,
                backend_id=backend_subnet['name'],
                defaults=dict(
                    name=backend_subnet['name'],
                    cidr=backend_subnet['cidr'],
                    gateway_ip=gateway_ip,
                    allocation_pools=backend_subnet['allocation_pools'],
                    dns_nameservers=backend_subnet['dns_nameservers'],
                )
            )
            self.pull_internal_ips(subnet, backend_subnet['ips'])

    def pull_internal_ips(self, subnet, internal_ips):
        for internal_ip in internal_ips:
            models.InternalIP.objects.update_or_create(
                settings=self.settings,
                subnet=subnet,
                backend_id=internal_ip['ip'],
                defaults={
                    'name': internal_ip['ip'],
                    'address': internal_ip['ip'],
                    'is_available': internal_ip['available'],
                }
            )



from django.conf import settings
from waldur_ansible.common import cache_utils as python_cache_utils
from waldur_ansible.jupyter_hub_management import models
from waldur_ansible.python_management.backend import locking_service as python_locking_service
from waldur_ansible.python_management.backend.locking_service import PythonManagementBackendLockBuilder

JUPYTOR_HUB_MANAGEMENT_LOCK = 'waldur_jupyter_hub_management_global_%s'


class JupyterHubConfigRequestProcessingFinishedLockingHandler(object):
    def handle_on_processing_finished(self, request):
        python_cache_utils.release_task_status(
            JupyterHubManagementBackendLockBuilder.build_global_lock(request.jupyter_hub_management.python_management))


class JupyterHubConfigRelatedToVirtualEnvRequestProcessingFinishedLockingHandler(object):
    def handle_on_processing_finished(self, request):
        python_cache_utils.release_task_status(
            python_locking_service.PythonManagementBackendLockBuilder.build_related_to_virt_env_lock(
                request.jupyter_hub_management.python_management, request.virtual_env_name))


class JupyterHubConfigProcessingAllowedDecider(object):
    def is_processing_allowed(self, request):
        global_lock = JupyterHubManagementBackendLockBuilder.build_global_lock(request.jupyter_hub_management.python_management)
        return not python_cache_utils.is_syncing(global_lock)


class JupyterHubRelatedToVirtualEnvProcessingAllowedDecider(object):
    def is_processing_allowed(self, request):
        virtual_env_lock = python_locking_service.PythonManagementBackendLockBuilder.build_related_to_virt_env_lock(
            request.jupyter_hub_management.python_management, request.virtual_env_name)
        global_lock = python_locking_service.PythonManagementBackendLockBuilder.build_global_lock(request.jupyter_hub_management.python_management)
        return not python_cache_utils.is_syncing(virtual_env_lock) and not python_cache_utils.is_syncing(global_lock)


class JupyterHubConfigSynchronizer(object):
    def lock(self, request):
        global_lock = JupyterHubManagementBackendLockBuilder.build_global_lock(request.jupyter_hub_management.python_management)
        python_cache_utils.renew_task_status(global_lock, settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_REQUEST_TIMEOUT'])


class JupyterHubRelatedToVirtualEnvSynchronizer(object):
    def lock(self, request):
        virtual_env_lock = PythonManagementBackendLockBuilder.build_related_to_virt_env_lock(
            request.jupyter_hub_management.python_management, request.virtual_env_name)
        python_cache_utils.renew_task_status(virtual_env_lock, settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_REQUEST_TIMEOUT'])


class JupyterHubManagementBackendLockingService(object):

    @staticmethod
    def lock_for_processing(request):
        synchronizer = JupyterHubManagementBackendLockBuilder.intantiate_synchronizer(type(request))
        return synchronizer.lock(request)

    @staticmethod
    def is_processing_allowed(request):
        processing_allowed_decider = JupyterHubManagementBackendLockBuilder.intantiate_processing_allowed_decider(type(request))
        return processing_allowed_decider.is_processing_allowed(request)

    @staticmethod
    def handle_on_processing_finished(request):
        locking_handler = JupyterHubManagementBackendLockBuilder.intantiate_processing_finished_locking_handler(type(request))
        locking_handler.handle_on_processing_finished(request)


class JupyterHubManagementBackendLockBuilder(object):
    REQUEST_TYPES_PROCESSING_ALLOWED_DECIDER = {
        models.JupyterHubManagementSyncConfigurationRequest: JupyterHubConfigProcessingAllowedDecider,
        models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest: JupyterHubRelatedToVirtualEnvProcessingAllowedDecider,
        models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest: JupyterHubRelatedToVirtualEnvProcessingAllowedDecider,
        models.JupyterHubManagementDeleteRequest: JupyterHubConfigProcessingAllowedDecider,
    }
    REQUEST_TYPES_SYNCHRONIZERS = {
        models.JupyterHubManagementSyncConfigurationRequest: JupyterHubConfigSynchronizer,
        models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest: JupyterHubRelatedToVirtualEnvSynchronizer,
        models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest: JupyterHubRelatedToVirtualEnvSynchronizer,
        models.JupyterHubManagementDeleteRequest: JupyterHubConfigSynchronizer,
    }
    REQUEST_TYPES_PROCESSING_FINISHED_LOCKING_HANDLER = {
        models.JupyterHubManagementSyncConfigurationRequest: JupyterHubConfigRequestProcessingFinishedLockingHandler,
        models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest: JupyterHubConfigRelatedToVirtualEnvRequestProcessingFinishedLockingHandler,
        models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest: JupyterHubConfigRelatedToVirtualEnvRequestProcessingFinishedLockingHandler,
        models.JupyterHubManagementDeleteRequest: JupyterHubConfigRequestProcessingFinishedLockingHandler,
    }

    @staticmethod
    def intantiate_processing_allowed_decider(jupyter_hub_management_request_class):
        processing_allowed_decider_class = JupyterHubManagementBackendLockBuilder.REQUEST_TYPES_PROCESSING_ALLOWED_DECIDER \
            .get(jupyter_hub_management_request_class)
        return processing_allowed_decider_class()

    @staticmethod
    def intantiate_processing_finished_locking_handler(jupyter_hub_management_request_class):
        locking_handler_class = JupyterHubManagementBackendLockBuilder.REQUEST_TYPES_PROCESSING_FINISHED_LOCKING_HANDLER \
            .get(jupyter_hub_management_request_class)
        return locking_handler_class()

    @staticmethod
    def intantiate_synchronizer(jupyter_hub_management_request_class):
        locking_handler_class = JupyterHubManagementBackendLockBuilder.REQUEST_TYPES_SYNCHRONIZERS \
            .get(jupyter_hub_management_request_class)
        return locking_handler_class()

    @staticmethod
    def build_global_lock(persisted_jupyter_hub_management):
        return JUPYTOR_HUB_MANAGEMENT_LOCK % persisted_jupyter_hub_management.pk



from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from rest_framework import serializers


User = get_user_model()


class AuthSerializer(serializers.Serializer):
    client_id = serializers.CharField()
    redirect_uri = serializers.CharField()
    code = serializers.CharField()


class RegistrationSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('username', 'full_name', 'email', 'password')
        extra_kwargs = {
            'password': {'write_only': True},
        }

    def validate_email(self, value):
        if User.objects.filter(email=value).exists():
            raise serializers.ValidationError('User with this email already exists')
        return value

    def create(self, validated_data):
        return User.objects.create_user(**validated_data)


class ActivationSerializer(serializers.Serializer):
    user_uuid = serializers.CharField()
    token = serializers.CharField()

    def validate(self, attrs):
        try:
            self.user = User.objects.get(uuid=attrs['user_uuid'], is_active=False)
        except User.DoesNotExist:
            raise serializers.ValidationError('Invalid user UUID')

        if not default_token_generator.check_token(self.user, attrs['token']):
            raise serializers.ValidationError('Invalid token')
        return attrs



from django.contrib import admin
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import admin as core_admin

from . import models, tasks


class ProfileAdmin(core_admin.ExtraActionsMixin, admin.ModelAdmin):
    list_display = ('username', 'user', 'is_active', 'agreement_date')
    readonly_fields = ('username', 'user', 'is_active', 'agreement_date')
    list_filter = ('is_active', )
    search_fields = ('username', )

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_extra_actions(self):
        return [
            self.sync_groups,
            self.sync_names,
            self.sync_gecos,
        ]

    def sync_groups(self, request):
        tasks.schedule_sync()
        self.message_user(request, _('Groups synchronization has been scheduled.'))
        return redirect(reverse('admin:waldur_freeipa_profile_changelist'))

    def sync_names(self, request):
        tasks.schedule_sync_names()
        self.message_user(request, _('Names synchronization has been scheduled.'))
        return redirect(reverse('admin:waldur_freeipa_profile_changelist'))

    def sync_gecos(self, request):
        tasks.schedule_sync_gecos()
        self.message_user(request, _('GECOS synchronization has been scheduled.'))
        return redirect(reverse('admin:waldur_freeipa_profile_changelist'))


admin.site.register(models.Profile, ProfileAdmin)



from __future__ import unicode_literals

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from model_utils.models import TimeStampedModel

from waldur_ansible.common import models as common_models
from waldur_ansible.python_management import models as python_management_models
from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models

User = get_user_model()


class JupyterHubOAuthType(object):
    # these numbers are important for both frontend and ansible playbook logic
    GITLAB = 1
    AZURE = 2

    CHOICES = (
        (GITLAB, 'GitLab'),
        (AZURE, 'Microsoft Azure'),
    )


class JupyterHubOAuthConfig(common_models.UuidStrMixin):
    type = models.PositiveSmallIntegerField(choices=JupyterHubOAuthType.CHOICES)

    oauth_callback_url = models.CharField(max_length=255)
    client_id = models.CharField(max_length=255)
    client_secret = models.CharField(max_length=255)
    # required for Azure:
    tenant_id = models.CharField(max_length=255, blank=True, null=True)
    gitlab_host = models.CharField(max_length=255, blank=True, null=True)


class JupyterHubManagement(common_models.UuidStrMixin,
                           TimeStampedModel,
                           common_models.ApplicationModel):
    user = models.ForeignKey(User, related_name='+')
    python_management = models.ForeignKey(python_management_models.PythonManagement, related_name='jupyter_hub_management')

    session_time_to_live_hours = models.IntegerField(default=24)
    jupyter_hub_oauth_config = models.OneToOneField(
        JupyterHubOAuthConfig,
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )

    instance_content_type = models.ForeignKey(ContentType, null=True, related_name='+')
    instance_object_id = models.PositiveIntegerField(null=True)
    instance = GenericForeignKey('instance_content_type', 'instance_object_id')

    project = models.ForeignKey(structure_models.Project, null=True, related_name='+')

    # holds reference to jupyter_hub_users

    class Meta:
        unique_together = ('instance_content_type', 'instance_object_id')

    class Permissions:
        project_path = 'project'
        customer_path = 'project__customer'

    @staticmethod
    def get_url_name():
        return 'jupyter_hub_management'

    def get_admin_users(self):
        return JupyterHubUser.objects.filter(jupyter_hub_management=self, admin=True)

    def get_whitelisted_users(self):
        return JupyterHubUser.objects.filter(jupyter_hub_management=self, whitelisted=True)


class JupyterHubUser(common_models.UuidStrMixin):
    jupyter_hub_management = models.ForeignKey(JupyterHubManagement, on_delete=models.CASCADE, related_name='jupyter_hub_users')
    admin = models.BooleanField(default=False)
    # plays role only in case of OAuth
    whitelisted = models.BooleanField(default=False)
    username = models.CharField(max_length=255)
    password = models.CharField(max_length=255, null=True, blank=True)


class JupyterHubManagementRequest(common_models.UuidStrMixin,
                                  core_models.StateMixin,
                                  TimeStampedModel,
                                  common_models.OutputMixin):
    jupyter_hub_management = models.ForeignKey(JupyterHubManagement, on_delete=models.CASCADE, related_name='+')

    class Meta(object):
        abstract = True

    def get_backend(self):
        from waldur_ansible.jupyter_hub_management.backend.backend import JupyterHubManagementBackend
        return JupyterHubManagementBackend()


class JupyterHubManagementSyncConfigurationRequest(JupyterHubManagementRequest):
    # holds make_virtual_env_global_requests reference
    pass


class JupyterHubManagementMakeVirtualEnvironmentGlobalRequest(
        python_management_models.VirtualEnvMixin,
        JupyterHubManagementRequest):
    update_configuration_request = models.ForeignKey(JupyterHubManagementSyncConfigurationRequest, related_name="make_virtual_env_global_requests", null=True)


class JupyterHubManagementDeleteRequest(JupyterHubManagementRequest):
    pass


class JupyterHubManagementMakeVirtualEnvironmentLocalRequest(
        python_management_models.VirtualEnvMixin,
        JupyterHubManagementRequest):
    pass



import django_filters

from waldur_core.core import filters as core_filters

from . import models


class PackageTemplateFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    service_settings = core_filters.URLFilter(
        view_name='servicesettings-detail', name='service_settings__uuid')
    service_settings_uuid = django_filters.UUIDFilter(name='service_settings__uuid')
    openstack_package_service_settings = core_filters.URLFilter(
        view_name='servicesettings-detail', name='openstack_packages__service_settings__uuid')
    openstack_package_service_settings_uuid = django_filters.UUIDFilter(
        name='openstack_packages__service_settings__uuid')

    class Meta(object):
        model = models.PackageTemplate
        fields = ['archived']


class OpenStackPackageFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(name='tenant__name', lookup_expr='icontains')
    customer = core_filters.URLFilter(
        view_name='customer-detail', name='tenant__service_project_link__project__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='tenant__service_project_link__project__customer__uuid')
    project = core_filters.URLFilter(
        view_name='project-detail', name='tenant__service_project_link__project__uuid')
    project_uuid = django_filters.UUIDFilter(name='tenant__service_project_link__project__uuid')
    template = core_filters.URLFilter(view_name='package-template-detail', name='template__uuid')
    template_uuid = django_filters.UUIDFilter(name='template__uuid')
    tenant = core_filters.URLFilter(view_name='openstack-tenant-detail', name='tenant__uuid')
    tenant_uuid = django_filters.UUIDFilter(name='tenant__uuid')
    service_settings = core_filters.URLFilter(
        view_name='servicesettings-detail', name='service_settings__uuid')
    service_settings_uuid = django_filters.UUIDFilter(name='service_settings__uuid')

    class Meta(object):
        model = models.OpenStackPackage
        fields = '__all__'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-06 17:45
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0050_reset_cloud_spl_quota_limits'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='agreement_number',
            field=models.PositiveIntegerField(blank=True, null=True, unique=True),
        ),
        migrations.AddField(
            model_name='customer',
            name='email',
            field=models.EmailField(blank=True, max_length=75, verbose_name='email address'),
        ),
        migrations.AddField(
            model_name='customer',
            name='phone_number',
            field=models.CharField(blank=True, max_length=255, verbose_name='phone number'),
        ),
    ]



from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.status import HTTP_202_ACCEPTED, HTTP_423_LOCKED, HTTP_412_PRECONDITION_FAILED
from waldur_ansible.common import cache_utils
from waldur_ansible.python_management.backend import locking_service

from waldur_core.core.models import StateMixin
from . import models, executors


class PythonManagementService(object):
    executor = executors.PythonManagementRequestExecutor

    @staticmethod
    def build_response(locked_virtual_envs):
        return {
            'locked': 'Python management is locked, please retry later',
            'global_lock': not bool(locked_virtual_envs),
            'locked_virtual_envs': locked_virtual_envs,
        }

    @staticmethod
    def build_python_management_locked_response(locked_virtual_envs=None):
        if not locked_virtual_envs:
            locked_virtual_envs = []
        return Response(
            PythonManagementService.build_response(locked_virtual_envs),
            status=HTTP_423_LOCKED)

    def schedule_python_management_removal(self, persisted_python_management):
        if persisted_python_management.jupyter_hub_management.all():
            raise APIException(code=HTTP_412_PRECONDITION_FAILED)
        delete_request = models.PythonManagementDeleteRequest(python_management=persisted_python_management)

        if not locking_service.PythonManagementBackendLockingService.is_processing_allowed(delete_request):
            raise APIException(code=HTTP_423_LOCKED)

        delete_request.save()
        self.executor.execute(delete_request, async=True)

    def schedule_virtual_environments_search(self, persisted_python_management):
        find_virtual_envs_request = models.PythonManagementFindVirtualEnvsRequest(python_management=persisted_python_management)

        if not locking_service.PythonManagementBackendLockingService.is_processing_allowed(find_virtual_envs_request):
            raise APIException(code=HTTP_423_LOCKED)

        find_virtual_envs_request.save()
        self.executor.execute(find_virtual_envs_request, async=True)
        return Response({'status': 'Find installed virtual environments process has been scheduled.'},
                        status=HTTP_202_ACCEPTED)

    def schedule_installed_libraries_search(self, persisted_python_management, virtual_env_name):
        find_installed_libraries_request = models.PythonManagementFindInstalledLibrariesRequest(
            python_management=persisted_python_management, virtual_env_name=virtual_env_name)

        if not locking_service.PythonManagementBackendLockingService.is_processing_allowed(find_installed_libraries_request):
            raise APIException(code=HTTP_423_LOCKED)

        find_installed_libraries_request.save()
        self.executor.execute(find_installed_libraries_request, async=True)
        return Response(
            {'status': 'Find installed libraries in virtual environment process has been scheduled.'},
            status=HTTP_202_ACCEPTED)

    def schedule_virtual_environments_update(self, all_transient_virtual_environments, persisted_python_management):
        persisted_virtual_environments = persisted_python_management.virtual_environments.all()
        virtual_environments_to_create, virtual_environments_to_change, removed_virtual_environments = \
            self.identify_changed_created_removed_envs(
                all_transient_virtual_environments, persisted_virtual_environments)

        if cache_utils.is_syncing(
                locking_service.PythonManagementBackendLockBuilder.build_global_lock(persisted_python_management)):
            raise APIException(code=HTTP_423_LOCKED)

        self.create_or_refuse_requests(
            persisted_python_management, removed_virtual_environments,
            virtual_environments_to_change, virtual_environments_to_create)

    def identify_changed_created_removed_envs(self, all_transient_virtual_environments, persisted_virtual_environments):

        removed_virtual_environments = []
        virtual_environments_to_create = []
        virtual_environments_to_change = []

        for virtual_environment in persisted_virtual_environments:
            libraries_to_remove = []
            libraries_to_install = []
            virtual_environment_name = virtual_environment.name
            corresponding_transient_virtual_environment = self.find_corresponding_transient_virtual_environment(
                virtual_environment_name, all_transient_virtual_environments)
            if not corresponding_transient_virtual_environment:
                removed_virtual_environments.append(virtual_environment)
            else:
                transient_libraries = corresponding_transient_virtual_environment['installed_libraries']
                persisted_libraries = virtual_environment.installed_libraries.all()

                for installed_library in persisted_libraries:
                    transient_library = self.find_corresponding_transient_library(
                        installed_library, transient_libraries)
                    if not transient_library:
                        libraries_to_remove.append(
                            {'name': installed_library.name, 'version': installed_library.version})

                for transient_library in transient_libraries:
                    persisted_library = self.find_corresponding_persisted_library(
                        transient_library, persisted_libraries)

                    if not persisted_library:
                        libraries_to_install.append(transient_library)

                if libraries_to_remove or libraries_to_install:
                    virtual_environments_to_change.append({
                        'name': virtual_environment_name,
                        'libraries_to_install': libraries_to_install,
                        'libraries_to_remove': libraries_to_remove})

        for transient_virtual_environment in all_transient_virtual_environments:
            persisted_virtual_environment = self.find_corresponding_persisted_virtual_environment(
                transient_virtual_environment['name'], persisted_virtual_environments)
            if not persisted_virtual_environment:
                virtual_environments_to_create.append(transient_virtual_environment)

        return virtual_environments_to_create, virtual_environments_to_change, removed_virtual_environments

    def is_global_request(self, request):
        return not request.virtual_env_name

    def is_request_executing(self, request):
        return request.state != StateMixin.States.OK and request.state != StateMixin.States.ERRED

    def find_corresponding_persisted_library(self, transient_library, persisted_libraries):
        for persisted_library in persisted_libraries:
            if persisted_library.name == transient_library['name'] and persisted_library.version == transient_library['version']:
                return persisted_library
        return None

    def find_corresponding_transient_library(self, persisted_library, transient_libraries):
        for transient_library in transient_libraries:
            if transient_library['name'] == persisted_library.name and transient_library['version'] == persisted_library.version:
                return transient_library
        return None

    def find_corresponding_persisted_virtual_environment(self, virtual_environment_name, persisted_virtual_environments):
        for persisted_virtual_environment in persisted_virtual_environments:
            if persisted_virtual_environment.name == virtual_environment_name:
                return persisted_virtual_environment
        return None

    def find_corresponding_transient_virtual_environment(self, virtual_environment_name,
                                                         transient_virtual_environments):
        for transient_virtual_environment in transient_virtual_environments:
            if transient_virtual_environment['name'] == virtual_environment_name:
                return transient_virtual_environment
        return None

    def create_or_refuse_requests(self, persisted_python_management, removed_virtual_environments,
                                  virtual_environments_to_change, virtual_environments_to_create):
        for virtual_environment_to_create in virtual_environments_to_create:
            sync_request = models.PythonManagementSynchronizeRequest(
                python_management=persisted_python_management,
                libraries_to_install=virtual_environment_to_create['installed_libraries'],
                virtual_env_name=virtual_environment_to_create['name'])

            self.create_or_refuse_request(sync_request)

        for removed_virtual_environment in removed_virtual_environments:
            delete_virt_env_request = models.PythonManagementDeleteVirtualEnvRequest(
                python_management=persisted_python_management,
                virtual_env_name=removed_virtual_environment.name)

            self.create_or_refuse_request(delete_virt_env_request)

        for virtual_environment_to_change in virtual_environments_to_change:
            sync_request = models.PythonManagementSynchronizeRequest(
                python_management=persisted_python_management,
                libraries_to_install=virtual_environment_to_change['libraries_to_install'],
                libraries_to_remove=virtual_environment_to_change['libraries_to_remove'],
                virtual_env_name=virtual_environment_to_change['name'])

            self.create_or_refuse_request(sync_request)

    def create_or_refuse_request(self, sync_request):
        if locking_service.PythonManagementBackendLockingService.is_processing_allowed(sync_request):
            sync_request.save()
            self.executor.execute(sync_request, async=True)



import json

import django_filters
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions as rf_exceptions

from waldur_core.core import filters as core_filters

from . import models


class ServiceProviderFilter(django_filters.FilterSet):
    customer = core_filters.URLFilter(view_name='customer-detail', name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='customer__uuid')

    class Meta(object):
        model = models.ServiceProvider
        fields = []


class OfferingFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    customer = core_filters.URLFilter(view_name='customer-detail', name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='customer__uuid')
    attributes = django_filters.CharFilter(name='attributes', method='filter_attributes')

    def filter_attributes(self, queryset, name, value):
        try:
            value = json.loads(value)
        except ValueError:
            raise rf_exceptions.ValidationError(_('Filter attribute is not valid.'))

        for k, v in value.items():
            queryset = queryset.filter(attributes__contains={k: v})
        return queryset

    class Meta(object):
        model = models.Offering
        fields = ['attributes']


class ScreenshotFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(name='offering__uuid')

    o = django_filters.OrderingFilter(fields=('name', 'created'))

    class Meta(object):
        model = models.Screenshots
        fields = []


class OrderFilter(django_filters.FilterSet):
    project = core_filters.URLFilter(view_name='project-detail', name='project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__uuid')

    o = django_filters.OrderingFilter(fields=('created', 'approved_at', 'total_cost', 'state'))

    class Meta(object):
        model = models.Order
        fields = ['state']


class ItemFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(name='offering__uuid')

    order = core_filters.URLFilter(view_name='marketplace-order-detail', name='order__uuid')
    order_uuid = django_filters.UUIDFilter(name='order__uuid')

    class Meta(object):
        model = models.OrderItem
        fields = []


class PlanFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(name='offering__uuid')

    class Meta(object):
        model = models.Plan
        fields = []



from __future__ import unicode_literals

from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions, status


class PriceEstimateLimitExceeded(exceptions.APIException):
    status_code = status.HTTP_400_BAD_REQUEST

    def __init__(self, price_estimate):
        super(PriceEstimateLimitExceeded, self).__init__()
        message = _('Price for %(scope_type)s "%(scope_name)s" is over limit. Required: %(required)s, limit: %(limit)s')
        context = {
            'scope_type': price_estimate.content_type,
            'scope_name': price_estimate.scope.name,
            'required': price_estimate.total,
            'limit': price_estimate.limit,
        }
        self.detail = message % context

    def __str__(self):
        return self.detail



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def initialize_backupschedule_project_links(apps, schema_editor):
    BackupSchedule = apps.get_model('openstack_tenant', 'BackupSchedule')
    for backup_schedule in BackupSchedule.objects.iterator():
        backup_schedule.service_project_link = backup_schedule.instance.service_project_link
        backup_schedule.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0012_backupschedule_resource'),
    ]

    operations = [
        migrations.RunPython(initialize_backupschedule_project_links),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:02
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('python_management', '0004_removed_virtual_env_field_from_global_requests'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pythonmanagementsynchronizerequest',
            name='libraries_to_install',
            field=waldur_core.core.fields.JSONField(blank=True, default=list, help_text='List of libraries to install'),
        ),
        migrations.AlterField(
            model_name='pythonmanagementsynchronizerequest',
            name='libraries_to_remove',
            field=waldur_core.core.fields.JSONField(blank=True, default=list, help_text='List of libraries to remove'),
        ),
    ]



"""
TODO: drop patch when django-fsm package is updated.

If model with FSM state field has other fields that access their field value
via a property or a virtual Field, then creation of instances will fail.

There is pending patch in upstream project:
https://github.com/kmmbvnr/django-fsm/pull/171
"""

__all__ = ['monkey_patch_fields']


def subfield_get(self, obj, type=None):
    """
    Verbatim copy from:
    https://github.com/django/django/blob/1.9.13/django/db/models/fields/subclassing.py#L38
    """
    if obj is None:
        return self
    return obj.__dict__[self.field.name]


def get_field_name(self):
    return self.field.name


def patch_field_descriptor(cls):
    cls.__get__ = subfield_get
    setattr(cls, 'field_name', property(get_field_name))


def monkey_patch_fields():
    from django_fsm import FSMFieldDescriptor
    patch_field_descriptor(FSMFieldDescriptor)



from __future__ import unicode_literals

import os

import six
from PIL import Image
from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage as storage


def create_screenshot_thumbnail(screenshot):
    pic = screenshot.image
    fh = storage.open(pic.name, 'r')
    image = Image.open(fh)
    image.thumbnail(settings.WALDUR_MARKETPLACE['THUMBNAIL_SIZE'], Image.ANTIALIAS)
    fh.close()

    thumb_extension = os.path.splitext(pic.name)[1]
    thumb_extension = thumb_extension.lower()
    thumb_name = os.path.basename(pic.name)

    if thumb_extension in ['.jpg', '.jpeg']:
        FTYPE = 'JPEG'
    elif thumb_extension == '.gif':
        FTYPE = 'GIF'
    elif thumb_extension == '.png':
        FTYPE = 'PNG'
    else:
        return

    temp_thumb = six.StringIO()
    image.save(temp_thumb, FTYPE)
    temp_thumb.seek(0)
    screenshot.thumbnail.save(thumb_name, ContentFile(temp_thumb.read()), save=True)
    temp_thumb.close()



from __future__ import unicode_literals

import datetime
from functools import reduce
import itertools

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.core.validators import MaxLengthValidator
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models, transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
from model_utils.fields import AutoCreatedField
from model_utils.models import TimeStampedModel
import pyvat
from taggit.managers import TaggableManager

from waldur_core.core import fields as core_fields
from waldur_core.core import models as core_models
from waldur_core.core import utils as core_utils
from waldur_core.core.fields import JSONField
from waldur_core.core.models import CoordinatesMixin, AbstractFieldTracker
from waldur_core.core.validators import validate_name, validate_cidr_list, FileTypeValidator
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.monitoring.models import MonitoringModelMixin
from waldur_core.quotas import models as quotas_models, fields as quotas_fields
from waldur_core.structure import SupportedServices
from waldur_core.structure.images import ImageModelMixin
from waldur_core.structure.managers import StructureManager, filter_queryset_for_user, \
    ServiceSettingsManager, PrivateServiceSettingsManager, SharedServiceSettingsManager
from waldur_core.structure.signals import structure_role_granted, structure_role_revoked
from waldur_core.structure.utils import get_coordinates_by_ip, sort_dependencies


def validate_service_type(service_type):
    from django.core.exceptions import ValidationError
    if not SupportedServices.has_service_type(service_type):
        raise ValidationError(_('Invalid service type.'))


class StructureModel(models.Model):
    """ Generic structure model.
        Provides transparent interaction with base entities and relations like customer.
    """

    objects = StructureManager()

    class Meta(object):
        abstract = True

    def __getattr__(self, name):
        # add additional properties to the object according to defined Permissions class
        fields = ('customer', 'project')
        if name in fields:
            try:
                path = getattr(self.Permissions, name + '_path')
            except AttributeError:
                pass
            else:
                if not path == 'self' and '__' in path:
                    return reduce(getattr, path.split('__'), self)

        raise AttributeError(
            "'%s' object has no attribute '%s'" % (self._meta.object_name, name))


class StructureLoggableMixin(LoggableMixin):

    @classmethod
    def get_permitted_objects_uuids(cls, user):
        """
        Return query dictionary to search objects available to user.
        """
        uuids = filter_queryset_for_user(cls.objects.all(), user).values_list('uuid', flat=True)
        key = core_utils.camel_case_to_underscore(cls.__name__) + '_uuid'
        return {key: uuids}


class TagMixin(models.Model):
    """
    Add tags field and manage cache for tags.
    """
    class Meta:
        abstract = True

    tags = TaggableManager(related_name='+', blank=True)

    def get_tags(self):
        key = self._get_tag_cache_key()
        tags = cache.get(key)
        if tags is None:
            tags = list(self.tags.all().values_list('name', flat=True))
            cache.set(key, tags)
        return tags

    def clean_tag_cache(self):
        key = self._get_tag_cache_key()
        cache.delete(key)

    def _get_tag_cache_key(self):
        return 'tags:%s' % core_utils.serialize_instance(self)


class VATException(Exception):
    pass


class VATMixin(models.Model):
    """
    Add country, VAT number fields and check results from EU VAT Information Exchange System.
    Allows to compute VAT charge rate.
    """
    class Meta(object):
        abstract = True

    vat_code = models.CharField(max_length=20, blank=True, help_text=_('VAT number'))
    vat_name = models.CharField(max_length=255, blank=True,
                                help_text=_('Optional business name retrieved for the VAT number.'))
    vat_address = models.CharField(max_length=255, blank=True,
                                   help_text=_('Optional business address retrieved for the VAT number.'))

    is_company = models.BooleanField(default=False, help_text=_('Is company or private person'))
    country = core_fields.CountryField(blank=True)

    def get_vat_rate(self):
        charge = self.get_vat_charge()
        if charge.action == pyvat.VatChargeAction.charge:
            return charge.rate

        # Return None, if reverse_charge or no_charge action is applied

    def get_vat_charge(self):
        if not self.country:
            raise VATException(_('Unable to get VAT charge because buyer country code is not specified.'))

        seller_country = settings.WALDUR_CORE.get('SELLER_COUNTRY_CODE')
        if not seller_country:
            raise VATException(_('Unable to get VAT charge because seller country code is not specified.'))

        return pyvat.get_sale_vat_charge(
            datetime.date.today(),
            pyvat.ItemType.generic_electronic_service,
            pyvat.Party(self.country, self.is_company and self.vat_code),
            pyvat.Party(seller_country, True)
        )


class BasePermission(models.Model):
    class Meta(object):
        abstract = True

    user = models.ForeignKey(settings.AUTH_USER_MODEL, db_index=True)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, related_name='+')
    created = AutoCreatedField()
    expiration_time = models.DateTimeField(null=True, blank=True)
    is_active = models.NullBooleanField(default=True, db_index=True)

    @classmethod
    def get_url_name(cls):
        raise NotImplementedError

    @classmethod
    def get_expired(cls):
        return cls.objects.filter(expiration_time__lt=timezone.now(), is_active=True)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    def revoke(self):
        raise NotImplementedError


class PermissionMixin(object):
    """
    Base permission management mixin for customer and project.
    It is expected that reverse `permissions` relation is created for this model.
    Provides method to grant, revoke and check object permissions.
    """

    def has_user(self, user, role=None, timestamp=False):
        """
        Checks whether user has role in entity.
        `timestamp` can have following values:
            - False - check whether user has role in entity at the moment.
            - None - check whether user has permanent role in entity.
            - Datetime object - check whether user will have role in entity at specific timestamp.
        """
        permissions = self.permissions.filter(user=user, is_active=True)

        if role is not None:
            permissions = permissions.filter(role=role)

        if timestamp is None:
            permissions = permissions.filter(expiration_time=None)
        elif timestamp:
            permissions = permissions.filter(Q(expiration_time=None) | Q(expiration_time__gte=timestamp))

        return permissions.exists()

    @transaction.atomic()
    def add_user(self, user, role, created_by=None, expiration_time=None):
        permission = self.permissions.filter(user=user, role=role, is_active=True).first()
        if permission:
            return permission, False

        permission = self.permissions.create(
            user=user,
            role=role,
            is_active=True,
            created_by=created_by,
            expiration_time=expiration_time,
        )

        structure_role_granted.send(
            sender=self.__class__,
            structure=self,
            user=user,
            role=role,
            created_by=created_by,
        )

        return permission, True

    @transaction.atomic()
    def remove_user(self, user, role=None, removed_by=None):
        permissions = self.permissions.all().filter(user=user, is_active=True)

        if role is not None:
            permissions = permissions.filter(role=role)

        affected_permissions = list(permissions)
        permissions.update(is_active=None, expiration_time=timezone.now())

        for permission in affected_permissions:
            self.log_role_revoked(permission, removed_by)

    @transaction.atomic()
    def remove_all_users(self):
        for permission in self.permissions.all().iterator():
            permission.delete()
            self.log_role_revoked(permission)

    def log_role_revoked(self, permission, removed_by=None):
        structure_role_revoked.send(
            sender=self.__class__,
            structure=self,
            user=permission.user,
            role=permission.role,
            removed_by=removed_by,
        )


class CustomerRole(models.CharField):
    OWNER = 'owner'
    SUPPORT = 'support'

    CHOICES = (
        (OWNER, 'Owner'),
        (SUPPORT, 'Support'),
    )

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 30
        kwargs['choices'] = self.CHOICES
        super(CustomerRole, self).__init__(*args, **kwargs)


@python_2_unicode_compatible
class CustomerPermission(BasePermission):
    class Meta(object):
        unique_together = ('customer', 'role', 'user', 'is_active')

    class Permissions(object):
        customer_path = 'customer'

    customer = models.ForeignKey('structure.Customer', verbose_name=_('organization'), related_name='permissions')
    role = CustomerRole(db_index=True)
    tracker = FieldTracker(fields=['expiration_time'])

    @classmethod
    def get_url_name(cls):
        return 'customer_permission'

    def revoke(self):
        self.customer.remove_user(self.user, self.role)

    def __str__(self):
        return '%s | %s' % (self.customer.name, self.get_role_display())


def get_next_agreement_number():
    inital_number = settings.WALDUR_CORE['INITIAL_CUSTOMER_AGREEMENT_NUMBER']
    last_number = Customer.objects.aggregate(models.Max('agreement_number')).get('agreement_number__max')
    return (last_number or inital_number) + 1


@python_2_unicode_compatible
class Customer(core_models.UuidMixin,
               core_models.NameMixin,
               core_models.DescendantMixin,
               quotas_models.ExtendableQuotaModelMixin,
               PermissionMixin,
               VATMixin,
               LoggableMixin,
               ImageModelMixin,
               TimeStampedModel,
               StructureModel):
    class Permissions(object):
        customer_path = 'self'
        project_path = 'projects'

    native_name = models.CharField(max_length=160, default='', blank=True)
    abbreviation = models.CharField(max_length=12, blank=True)
    contact_details = models.TextField(blank=True, validators=[MaxLengthValidator(500)])
    agreement_number = models.PositiveIntegerField(null=True, blank=True, unique=True)
    email = models.EmailField(_('email address'), max_length=75, blank=True)
    phone_number = models.CharField(_('phone number'), max_length=255, blank=True)
    access_subnets = models.TextField(validators=[validate_cidr_list], blank=True, default='',
                                      help_text=_('Enter a comma separated list of IPv4 or IPv6 '
                                                  'CIDR addresses from where connection to self-service is allowed.'))
    registration_code = models.CharField(max_length=160, default='', blank=True)

    type = models.CharField(blank=True, max_length=150)
    address = models.CharField(blank=True, max_length=300)
    postal = models.CharField(blank=True, max_length=20)
    bank_name = models.CharField(blank=True, max_length=150)
    bank_account = models.CharField(blank=True, max_length=50)
    accounting_start_date = models.DateTimeField(_('Start date of accounting'), default=timezone.now)
    default_tax_percent = models.DecimalField(default=0, max_digits=4, decimal_places=2,
                                              validators=[MinValueValidator(0), MaxValueValidator(100)])
    tracker = FieldTracker()

    class Meta(object):
        verbose_name = _('organization')

    GLOBAL_COUNT_QUOTA_NAME = 'nc_global_customer_count'

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        enable_fields_caching = False
        nc_project_count = quotas_fields.CounterQuotaField(
            target_models=lambda: [Project],
            path_to_scope='customer',
        )
        nc_service_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Service.get_all_models(),
            path_to_scope='customer',
        )
        nc_service_project_link_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ServiceProjectLink.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_user_count = quotas_fields.QuotaField()
        nc_resource_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ResourceMixin.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_app_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ApplicationMixin.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_vm_count = quotas_fields.CounterQuotaField(
            target_models=lambda: VirtualMachine.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_private_cloud_count = quotas_fields.CounterQuotaField(
            target_models=lambda: PrivateCloud.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_storage_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Storage.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_volume_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='customer',
            target_field='size',
        )
        nc_snapshot_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Snapshot.get_all_models(),
            path_to_scope='customer',
            target_field='size',
        )

    def get_log_fields(self):
        return ('uuid', 'name', 'abbreviation', 'contact_details')

    def get_owners(self):
        return get_user_model().objects.filter(
            customerpermission__customer=self,
            customerpermission__is_active=True,
            customerpermission__role=CustomerRole.OWNER,
        )

    def get_support_users(self):
        return get_user_model().objects.filter(
            customerpermission__customer=self,
            customerpermission__is_active=True,
            customerpermission__role=CustomerRole.SUPPORT,
        )

    def get_users(self):
        """ Return all connected to customer users """
        return get_user_model().objects.filter(
            Q(customerpermission__customer=self,
              customerpermission__is_active=True) |
            Q(projectpermission__project__customer=self,
              projectpermission__is_active=True)
        ).distinct().order_by('username')

    def can_user_update_quotas(self, user):
        return user.is_staff

    def can_manage_role(self, user, role=None, timestamp=False):
        """
        Checks whether user can grant/update/revoke customer permissions.
        `timestamp` can have following values:
            - False - check whether user can manage permissions at the moment.
            - None - check whether user can permanently manage permissions.
            - Datetime object - check whether user will be able to manage permissions at specific timestamp.
        """
        return user.is_staff or (
            self.has_user(user, CustomerRole.OWNER, timestamp) and
            settings.WALDUR_CORE['OWNERS_CAN_MANAGE_OWNERS']
        )

    def get_children(self):
        return itertools.chain.from_iterable(
            m.objects.filter(customer=self) for m in [Project] + Service.get_all_models())

    def is_billable(self):
        return timezone.now() >= self.accounting_start_date

    @classmethod
    def get_permitted_objects_uuids(cls, user):
        if user.is_staff:
            customer_queryset = cls.objects.all()
        else:
            customer_queryset = cls.objects.filter(
                permissions__user=user,
                permissions__role=CustomerRole.OWNER,
                permissions__is_active=True
            )
        return {'customer_uuid': filter_queryset_for_user(customer_queryset, user).values_list('uuid', flat=True)}

    def __str__(self):
        return '%(name)s (%(abbreviation)s)' % {
            'name': self.name,
            'abbreviation': self.abbreviation
        }


class ProjectRole(models.CharField):
    ADMINISTRATOR = 'admin'
    MANAGER = 'manager'
    SUPPORT = 'support'

    CHOICES = (
        (ADMINISTRATOR, 'Administrator'),
        (MANAGER, 'Manager'),
        (SUPPORT, 'Support'),
    )

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 30
        kwargs['choices'] = self.CHOICES
        super(ProjectRole, self).__init__(*args, **kwargs)


@python_2_unicode_compatible
class ProjectPermission(core_models.UuidMixin, BasePermission):
    class Meta(object):
        unique_together = ('project', 'role', 'user', 'is_active')

    class Permissions(object):
        customer_path = 'project__customer'
        project_path = 'project'

    project = models.ForeignKey('structure.Project', related_name='permissions')
    role = ProjectRole(db_index=True)
    tracker = FieldTracker(fields=['expiration_time'])

    @classmethod
    def get_url_name(cls):
        return 'project_permission'

    def revoke(self):
        self.project.remove_user(self.user, self.role)

    def __str__(self):
        return '%s | %s' % (self.project.name, self.get_role_display())


@python_2_unicode_compatible
class ProjectType(core_models.DescribableMixin, core_models.UuidMixin, core_models.NameMixin):
    class Meta(object):
        verbose_name = _('Project type')
        verbose_name_plural = _('Project types')
        ordering = ['name']

    @classmethod
    def get_url_name(cls):
        return 'project_type'

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class Project(core_models.DescribableMixin,
              core_models.UuidMixin,
              core_models.NameMixin,
              core_models.DescendantMixin,
              quotas_models.ExtendableQuotaModelMixin,
              PermissionMixin,
              StructureLoggableMixin,
              TimeStampedModel,
              StructureModel):
    class Permissions(object):
        customer_path = 'customer'
        project_path = 'self'

    GLOBAL_COUNT_QUOTA_NAME = 'nc_global_project_count'

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        enable_fields_caching = False
        nc_resource_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ResourceMixin.get_all_models(),
            path_to_scope='project',
        )
        nc_app_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ApplicationMixin.get_all_models(),
            path_to_scope='project',
        )
        nc_vm_count = quotas_fields.CounterQuotaField(
            target_models=lambda: VirtualMachine.get_all_models(),
            path_to_scope='project',
        )
        nc_private_cloud_count = quotas_fields.CounterQuotaField(
            target_models=lambda: PrivateCloud.get_all_models(),
            path_to_scope='project',
        )
        nc_storage_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Storage.get_all_models(),
            path_to_scope='project',
        )
        nc_volume_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='project',
        )
        nc_snapshot_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Snapshot.get_all_models(),
            path_to_scope='project',
        )
        nc_service_project_link_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ServiceProjectLink.get_all_models(),
            path_to_scope='project',
        )
        nc_volume_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='project',
            target_field='size',
        )
        nc_snapshot_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Snapshot.get_all_models(),
            path_to_scope='project',
            target_field='size',
        )

    certifications = models.ManyToManyField(to='ServiceCertification', related_name='projects', blank=True)
    customer = models.ForeignKey(
        Customer, verbose_name=_('organization'), related_name='projects', on_delete=models.PROTECT)
    tracker = FieldTracker()
    type = models.ForeignKey(
        ProjectType, verbose_name=_('project type'), blank=True, null=True, on_delete=models.PROTECT)

    @property
    def full_name(self):
        return self.name

    def get_users(self, role=None):
        query = Q(
            projectpermission__project=self,
            projectpermission__is_active=True,
        )
        if role:
            query = query & Q(projectpermission__role=role)

        return get_user_model().objects.filter(query).order_by('username')

    def __str__(self):
        return '%(name)s | %(customer)s' % {
            'name': self.name,
            'customer': self.customer.name
        }

    def can_user_update_quotas(self, user):
        return user.is_staff or self.customer.has_user(user, CustomerRole.OWNER)

    def can_manage_role(self, user, role, timestamp=False):
        """
        Checks whether user can grant/update/revoke project permissions for specific role.
        `timestamp` can have following values:
            - False - check whether user can manage permissions at the moment.
            - None - check whether user can permanently manage permissions.
            - Datetime object - check whether user will be able to manage permissions at specific timestamp.
        """
        if user.is_staff:
            return True
        if self.customer.has_user(user, CustomerRole.OWNER, timestamp):
            return True

        return role == ProjectRole.ADMINISTRATOR and self.has_user(user, ProjectRole.MANAGER, timestamp)

    def get_log_fields(self):
        return ('uuid', 'customer', 'name')

    def get_parents(self):
        return [self.customer]

    def get_children(self):
        """
        Get all service project links connected to current project
        """
        return itertools.chain.from_iterable(
            m.objects.filter(project=self) for m in ServiceProjectLink.get_all_models())


@python_2_unicode_compatible
class ServiceCertification(core_models.UuidMixin, core_models.DescribableMixin):
    link = models.URLField(max_length=255, blank=True)
    # NameMixin is not used here as name has to be unique.
    name = models.CharField(_('name'), max_length=150, validators=[validate_name], unique=True)

    class Meta(object):
        verbose_name = 'Service Certification'
        verbose_name_plural = 'Service Certifications'
        ordering = ['-name']

    def __str__(self):
        return self.name

    @classmethod
    def get_url_name(cls):
        return 'service-certification'


@python_2_unicode_compatible
class ServiceSettings(quotas_models.ExtendableQuotaModelMixin,
                      core_models.UuidMixin,
                      core_models.NameMixin,
                      core_models.StateMixin,
                      TagMixin,
                      LoggableMixin):

    class Meta:
        verbose_name = "Service settings"
        verbose_name_plural = "Service settings"

    class Permissions(object):
        customer_path = 'customer'
        extra_query = dict(shared=True)

    customer = models.ForeignKey(Customer,
                                 verbose_name=_('organization'),
                                 related_name='service_settings',
                                 blank=True,
                                 null=True)
    backend_url = core_fields.BackendURLField(max_length=200, blank=True, null=True)
    username = models.CharField(max_length=100, blank=True, null=True)
    password = models.CharField(max_length=100, blank=True, null=True)
    domain = models.CharField(max_length=200, blank=True, null=True)
    token = models.CharField(max_length=255, blank=True, null=True)
    certificate = models.FileField(upload_to='certs', blank=True, null=True,
                                   validators=[FileTypeValidator(
                                       allowed_types=[
                                           'application/x-pem-file',
                                           'application/x-x509-ca-cert',
                                           'text/plain'],
                                       allowed_extensions=['pem'])])
    type = models.CharField(max_length=255, db_index=True, validators=[validate_service_type])
    options = JSONField(default=dict, help_text=_('Extra options'), blank=True)
    geolocations = JSONField(default=list, blank=True,
                             help_text=_('List of latitudes and longitudes. For example: '
                                         '[{"latitude": 123, "longitude": 345}, {"latitude": 456, "longitude": 678}]'))
    shared = models.BooleanField(default=False, help_text=_('Anybody can use it'))
    homepage = models.URLField(max_length=255, blank=True)
    terms_of_services = models.URLField(max_length=255, blank=True)
    certifications = models.ManyToManyField(to='ServiceCertification', related_name='service_settings', blank=True)

    tracker = FieldTracker()

    # service settings scope - VM that contains service
    content_type = models.ForeignKey(ContentType, null=True)
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')

    objects = ServiceSettingsManager('scope')

    def get_backend(self, **kwargs):
        return SupportedServices.get_service_backend(self.type)(self, **kwargs)

    def get_option(self, name):
        options = self.options or {}
        if name in options:
            return options.get(name)
        else:
            defaults = self.get_backend().DEFAULTS
            return defaults.get(name)

    def __str__(self):
        return '%s (%s)' % (self.name, self.get_type_display())

    def get_log_fields(self):
        return ('uuid', 'name', 'customer')

    def _get_log_context(self, entity_name):
        context = super(ServiceSettings, self)._get_log_context(entity_name)
        context['service_settings_type'] = self.get_type_display()
        return context

    def get_type_display(self):
        return SupportedServices.get_name_for_type(self.type)

    def get_services(self):
        service_model = SupportedServices.get_service_models()[self.type]['service']
        return service_model.objects.filter(settings=self)

    def unlink_descendants(self):
        for service in self.get_services():
            service.unlink_descendants()
            service.delete()


class SharedServiceSettings(ServiceSettings):
    """Required for a clear separation of shared/unshared service settings on admin."""

    objects = SharedServiceSettingsManager()

    class Meta(object):
        proxy = True
        verbose_name_plural = _('Shared provider settings')


class PrivateServiceSettings(ServiceSettings):
    """Required for a clear separation of shared/unshared service settings on admin."""

    objects = PrivateServiceSettingsManager()

    class Meta(object):
        proxy = True
        verbose_name_plural = _('Private provider settings')


@python_2_unicode_compatible
class Service(core_models.UuidMixin,
              core_models.DescendantMixin,
              quotas_models.QuotaModelMixin,
              LoggableMixin,
              StructureModel):
    """ Base service class. """

    class Meta(object):
        abstract = True
        unique_together = ('customer', 'settings')

    class Permissions(object):
        customer_path = 'customer'
        project_path = 'projects'

    settings = models.ForeignKey(ServiceSettings)
    customer = models.ForeignKey(Customer, verbose_name=_('organization'))
    available_for_all = models.BooleanField(
        default=False,
        help_text=_('Service will be automatically added to all customers projects if it is available for all')
    )
    projects = NotImplemented

    def __init__(self, *args, **kwargs):
        AbstractFieldTracker().finalize_class(self.__class__, 'tracker')
        super(Service, self).__init__(*args, **kwargs)

    def __str__(self):
        return self.settings.name

    def get_backend(self, **kwargs):
        return self.settings.get_backend(**kwargs)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return cls._meta.app_label

    def get_log_fields(self):
        return ('uuid', 'customer', 'settings')

    def _get_log_context(self, entity_name):
        context = super(Service, self)._get_log_context(entity_name)
        context['service_type'] = SupportedServices.get_name_for_model(self)
        return context

    def get_service_project_links(self):
        """
        Generic method for getting queryset of service project links related to current service
        """
        return self.projects.through.objects.filter(service=self)

    def get_parents(self):
        return [self.settings, self.customer]

    def get_children(self):
        return self.get_service_project_links()

    def unlink_descendants(self):
        descendants = sort_dependencies(self._meta.model, self.get_descendants())
        for descendant in descendants:
            if isinstance(descendant, ResourceMixin):
                descendant.unlink()
            descendant.delete()


class BaseServiceProperty(core_models.BackendModelMixin, core_models.UuidMixin, core_models.NameMixin, models.Model):
    """ Base service properties like image, flavor, region,
        which are usually used for Resource provisioning.
    """

    class Meta(object):
        abstract = True

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return '{}-{}'.format(cls._meta.app_label, cls.__name__.lower())

    @classmethod
    def get_backend_fields(cls):
        return super(BaseServiceProperty, cls).get_backend_fields() + ('backend_id', 'name')


@python_2_unicode_compatible
class ServiceProperty(BaseServiceProperty):

    class Meta(object):
        abstract = True
        unique_together = ('settings', 'backend_id')

    settings = models.ForeignKey(ServiceSettings, related_name='+')
    backend_id = models.CharField(max_length=255, db_index=True)

    def __str__(self):
        return '{0} | {1}'.format(self.name, self.settings)


@python_2_unicode_compatible
class GeneralServiceProperty(BaseServiceProperty):
    """
    Service property which is not connected to settings
    """

    class Meta(object):
        abstract = True

    backend_id = models.CharField(max_length=255, unique=True)

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class ServiceProjectLink(quotas_models.QuotaModelMixin,
                         core_models.DescendantMixin,
                         LoggableMixin,
                         StructureModel):
    """ Base service-project link class. See Service class for usage example. """

    class States(object):
        OK = 'OK'
        ERRED = 'ERRED'
        WARNING = 'WARNING'

        CHOICES = [OK, ERRED, WARNING]

    class Meta(object):
        abstract = True
        unique_together = ('service', 'project')

    class Permissions(object):
        customer_path = 'service__customer'
        project_path = 'project'

    service = NotImplemented
    project = models.ForeignKey(Project)

    def get_backend(self, **kwargs):
        return self.service.get_backend(**kwargs)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return cls._meta.app_label + '-spl'

    def get_log_fields(self):
        return ('project', 'service',)

    def get_parents(self):
        return [self.project, self.service]

    def get_children(self):
        resource_models = [m for m in ResourceMixin.get_all_models() + SubResource.get_all_models()
                           if m.service_project_link.field.related_model == self.__class__]
        return itertools.chain.from_iterable(
            m.objects.filter(service_project_link=self) for m in resource_models)

    @property
    def validation_state(self):
        """
        Defines whether a  service compliant with required project certifications.
        """
        if set(self.project.certifications.all()).issubset(set(self.service.settings.certifications.all())):
            return self.States.OK
        else:
            return self.States.ERRED

    @property
    def is_valid(self):
        return self.validation_state == self.States.OK

    @property
    def validation_message(self):
        """
        Validation result clarification.
        """
        if not self.is_valid:
            service_certifications = self.service.settings.certifications.all()
            project_certifications = self.project.certifications.all()
            missing_certifications = set(project_certifications) - set(service_certifications)
            return _('Provider does not match with project\'s security policy. Certifications are missing: "%s"') % ', '.join([c.name for c in missing_certifications])
        else:
            return ''

    def __str__(self):
        return '{0} | {1}'.format(self.service.settings.name, self.project.name)


class CloudServiceProjectLink(ServiceProjectLink):
    """
    Represents a link between a project and a cloud service that provides VPS or VPC (e.g. Amazon, DO, OpenStack).
    """

    class Meta(ServiceProjectLink.Meta):
        abstract = True

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        vcpu = quotas_fields.QuotaField()
        ram = quotas_fields.QuotaField()
        storage = quotas_fields.QuotaField()

    def can_user_update_quotas(self, user):
        return user.is_staff or self.service.customer.has_user(user, CustomerRole.OWNER)


class ApplicationMixin(models.Model):

    class Meta(object):
        abstract = True

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]


@python_2_unicode_compatible
class ResourceMixin(MonitoringModelMixin,
                    core_models.UuidMixin,
                    core_models.DescribableMixin,
                    core_models.NameMixin,
                    core_models.DescendantMixin,
                    core_models.BackendModelMixin,
                    LoggableMixin,
                    TagMixin,
                    TimeStampedModel,
                    StructureModel):

    """ Base resource class. Resource is a provisioned entity of a service,
        for example: a VM in OpenStack or AWS, or a repository in Github.
    """

    class Meta(object):
        abstract = True

    class Permissions(object):
        customer_path = 'service_project_link__project__customer'
        project_path = 'service_project_link__project'
        service_path = 'service_project_link__service'

    service_project_link = NotImplemented
    backend_id = models.CharField(max_length=255, blank=True)

    @classmethod
    def get_backend_fields(cls):
        return super(ResourceMixin, cls).get_backend_fields() + ('backend_id',)

    def get_backend(self, **kwargs):
        return self.service_project_link.get_backend(**kwargs)

    def get_cost(self, start_date, end_date):
        raise NotImplementedError(
            "Please refer to waldur_core.billing.tasks.debit_customers while implementing it")

    def get_access_url(self):
        # default behaviour. Override in subclasses if applicable
        return None

    def get_access_url_name(self):
        return None

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls) and not issubclass(model, SubResource)]

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return '{}-{}'.format(cls._meta.app_label, cls.__name__.lower())

    def get_log_fields(self):
        return ('uuid', 'name', 'service_project_link', 'full_name')

    @property
    def full_name(self):
        return '%s %s' % (SupportedServices.get_name_for_model(self).replace('.', ' '), self.name)

    def _get_log_context(self, entity_name):
        context = super(ResourceMixin, self)._get_log_context(entity_name)
        # XXX: Add resource_full_name here, because event context does not support properties as fields
        context['resource_full_name'] = self.full_name
        # required for lookups in ElasticSearch by the client
        context['resource_type'] = SupportedServices.get_name_for_model(self)

        # XXX: a hack for IaaS / PaaS / SaaS tags
        # XXX: should be moved to itacloud assembly
        if self.tags.filter(name='IaaS').exists():
            context['resource_delivery_model'] = 'IaaS'
        elif self.tags.filter(name='PaaS').exists():
            context['resource_delivery_model'] = 'PaaS'
        elif self.tags.filter(name='SaaS').exists():
            context['resource_delivery_model'] = 'SaaS'

        return context

    def filter_by_logged_object(self):
        return {
            'resource_uuid': self.uuid.hex,
            'resource_type': SupportedServices.get_name_for_model(self)
        }

    def get_parents(self):
        return [self.service_project_link]

    def __str__(self):
        return self.name

    def increase_backend_quotas_usage(self, validate=True):
        """ Increase usage of quotas that were consumed by resource on creation.

            If validate is True - method should raise QuotaValidationError if
            at least one of increased quotas if over limit.
        """
        pass

    def decrease_backend_quotas_usage(self):
        """ Decrease usage of quotas that were released on resource deletion """
        pass

    def unlink(self):
        # XXX: add special attribute to an instance in order to be tracked by signal handler
        setattr(self, 'PERFORM_UNLINK', True)

    @property
    def service_settings(self):
        return self.service_project_link.service.settings


# TODO: rename to Resource
class NewResource(ResourceMixin, core_models.StateMixin):

    class Meta(object):
        abstract = True


class VirtualMachine(CoordinatesMixin, core_models.RuntimeStateMixin, NewResource):

    def __init__(self, *args, **kwargs):
        AbstractFieldTracker().finalize_class(self.__class__, 'tracker')
        super(VirtualMachine, self).__init__(*args, **kwargs)

    cores = models.PositiveSmallIntegerField(default=0, help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(default=0, help_text=_('Memory size in MiB'))
    disk = models.PositiveIntegerField(default=0, help_text=_('Disk size in MiB'))
    min_ram = models.PositiveIntegerField(default=0, help_text=_('Minimum memory size in MiB'))
    min_disk = models.PositiveIntegerField(default=0, help_text=_('Minimum disk size in MiB'))

    image_name = models.CharField(max_length=150, blank=True)

    key_name = models.CharField(max_length=50, blank=True)
    key_fingerprint = models.CharField(max_length=47, blank=True)

    user_data = models.TextField(
        blank=True,
        help_text=_('Additional data that will be added to instance on provisioning'))
    start_time = models.DateTimeField(blank=True, null=True)

    class Meta(object):
        abstract = True

    def detect_coordinates(self):
        if self.external_ips:
            return get_coordinates_by_ip(self.external_ips)

    def get_access_url(self):
        if self.external_ips:
            return self.external_ips
        if self.internal_ips:
            return self.internal_ips
        return None

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @property
    def external_ips(self):
        """
        Returns a list of external IPs.
        Implementation of this method in all derived classes guarantees all virtual machine have the same interface.
        For instance:
         - SaltStack (aws) handles IPs as private and public IP addresses;
         - DigitalOcean has only 1 external ip called ip_address.
        """
        return []

    @property
    def internal_ips(self):
        """
        Returns a list of internal IPs.
        Implementation of this method in all derived classes guarantees all virtual machine have the same interface.
        For instance:
         - SaltStack (aws) handles IPs as private and public IP addresses;
         - DigitalOcean does not support internal IP at the moment.
        """
        return []


class PrivateCloud(quotas_models.QuotaModelMixin, core_models.RuntimeStateMixin, NewResource):
    extra_configuration = JSONField(default=dict,
                                    help_text=_('Configuration details that are not represented on backend.'))

    class Meta(object):
        abstract = True


class Storage(core_models.RuntimeStateMixin, NewResource):
    size = models.PositiveIntegerField(help_text=_('Size in MiB'))

    class Meta(object):
        abstract = True


class Volume(Storage):
    class Meta(object):
        abstract = True


class Snapshot(Storage):
    class Meta(object):
        abstract = True


class SubResource(NewResource):
    """ Resource dependent object that cannot exist without resource. """
    class Meta(object):
        abstract = True

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-13 09:18
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0002_resource'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('icon_url', models.URLField(blank=True, verbose_name='icon url')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name='project',
            name='template',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_jira.ProjectTemplate'),
            preserve_default=False,
        ),
    ]



from __future__ import unicode_literals

import collections

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from django.db.models import Q
from django.http import HttpResponse, Http404
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import decorators, exceptions, permissions, status, response, viewsets

from waldur_core.core import views as core_views
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import models as structure_models
from waldur_core.structure import permissions as structure_permissions
from waldur_core.structure import views as structure_views
from waldur_core.users import models as user_models
from waldur_core.users import tasks as user_tasks
from waldur_mastermind.support import backend as support_backend

from . import serializers, models, filters, quotas, tasks


def is_expert_manager(user):
    if user.is_staff:
        return True

    return models.ExpertProvider.objects.filter(
        customer__permissions__is_active=True,
        customer__permissions__user=user,
        customer__permissions__role=structure_models.CustomerRole.OWNER,
    ).exists()


def create_team_invitations(team, project, current_user):
    for permission in team.permissions.filter(is_active=True):
        invitation = user_models.Invitation.objects.create(
            created_by=current_user,
            email=permission.user.email,
            customer=project.customer,
            project=project,
            project_role=structure_models.ProjectRole.ADMINISTRATOR,
            link_template=settings.WALDUR_EXPERTS['INVITATION_LINK_TEMPLATE'],
        )
        username = current_user.full_name or current_user.username
        user_tasks.send_invitation.delay(invitation.uuid.hex, username)


def cancel_team_invitations(team, project):
    user_emails = [permission.user.email for permission in team.permissions.filter(is_active=True)]
    invitations = user_models.Invitation.objects.filter(
        email__in=user_emails,
        project=project,
        project_role=structure_models.ProjectRole.ADMINISTRATOR,
        state=user_models.Invitation.State.PENDING,
    )
    for invitation in invitations:
        invitation.cancel()


def revoke_request_permissions(expert_request):
    if not hasattr(expert_request, 'contract'):
        return
    team = expert_request.contract.team
    project = expert_request.project
    for permission in team.permissions.filter(is_active=True):
        project.remove_user(permission.user)
    cancel_team_invitations(team, project)


class ExpertProviderViewSet(viewsets.ModelViewSet):
    queryset = models.ExpertProvider.objects.all()
    serializer_class = serializers.ExpertProviderSerializer
    lookup_field = 'uuid'
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.ExpertProviderFilter


class ExpertRequestViewSet(core_views.ActionsViewSet):
    queryset = models.ExpertRequest.objects.all()
    serializer_class = serializers.ExpertRequestSerializer
    lookup_field = 'uuid'
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.ExpertRequestFilter
    disabled_actions = ['destroy']

    def get_queryset(self):
        qs = super(ExpertRequestViewSet, self).get_queryset()

        if self.request.user.is_staff:
            return qs
        elif is_expert_manager(self.request.user):
            qs = qs.filtered_for_manager(self.request.user)
        else:
            qs = qs.filtered_for_user(self.request.user)

        return qs

    @decorators.list_route()
    def configured(self, request):
        return response.Response(settings.WALDUR_EXPERTS['CONTRACT'], status=status.HTTP_200_OK)

    @transaction.atomic()
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        expert_request = serializer.save()
        support_backend.get_active_backend().create_issue(expert_request.issue)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    @transaction.atomic()
    @decorators.detail_route(methods=['post'])
    def cancel(self, request, *args, **kwargs):
        expert_request = self.get_object()
        expert_request.state = models.ExpertRequest.States.CANCELLED
        expert_request.save(update_fields=['state'])
        revoke_request_permissions(expert_request)
        return response.Response({'status': _('Expert request has been cancelled.')}, status=status.HTTP_200_OK)

    @transaction.atomic()
    @decorators.detail_route(methods=['post'])
    def complete(self, request, *args, **kwargs):
        expert_request = self.get_object()
        expert_request.state = models.ExpertRequest.States.COMPLETED
        expert_request.save(update_fields=['state'])
        revoke_request_permissions(expert_request)
        return response.Response({'status': _('Expert request has been completed.')}, status=status.HTTP_200_OK)

    @decorators.detail_route()
    def users(self, request, uuid=None):
        """
        Returns dict, where key is user UUID, and value is list of roles: staff, support, owner, expert.

        1. If user is an owner of organization that has issued current expert request -> owner.
        2. If user is from an organization that has submitted a bid -> expert.
        3. If user is staff/support -> staff, support.

        For example:
        {
            "0b02d56ebb0d4c6cb00a0728b5d9f349": ["owner"],
            "77020e89a4c54b189d0b27a2a863824f": ["expert"],
            "3d8ddca6c44a4cbda9a49e7c7cc1099b": ["staff", "support"]
        }
        """
        expert_request = self.get_object()
        roles = collections.defaultdict(list)

        if expert_request.issue:
            authors = list(expert_request.issue.comments.values_list('author__user__id', flat=True))
            users = get_user_model().objects.filter(id__in=authors)

            owners = expert_request.customer.get_owners()
            owners = [key.hex for key in owners.values_list('uuid', flat=True)]

            teams = list(expert_request.bids.values_list('team_id', flat=True))
            experts = get_user_model().objects.filter(
                projectpermission__project__in=teams,
                projectpermission__is_active=True
            )
            experts = [key.hex for key in experts.values_list('uuid', flat=True)]

            for user in users:
                key = user.uuid.hex
                if user.is_staff:
                    roles[key].append('staff')
                if user.is_support:
                    roles[key].append('support')
                if key in owners:
                    roles[key].append('owner')
                if key in experts:
                    roles[key].append('expert')

        return response.Response(status=status.HTTP_200_OK, data=dict(roles))

    @decorators.detail_route()
    def pdf(self, request, uuid=None):
        expert_request = self.get_object()

        try:
            contract = models.ExpertContract.objects.get(request=expert_request)
            if not contract.has_file():
                raise Http404()

            pdf_file = contract.get_file()
            file_response = HttpResponse(pdf_file, content_type='application/pdf')
            filename = contract.get_filename()
            file_response['Content-Disposition'] = 'attachment; filename="{filename}"'.format(filename=filename)
            return file_response
        except models.ExpertContract.DoesNotExist:
            raise Http404()

    def is_valid_request(request, view, obj=None):
        expert_request = obj

        if not expert_request:
            return

        if expert_request.state not in (models.ExpertRequest.States.ACTIVE, models.ExpertRequest.States.PENDING):
            raise exceptions.ValidationError(_('Expert request should be in active or pending state.'))

    def is_owner(request, view, obj=None):
        expert_request = obj

        if not expert_request:
            return

        if not structure_permissions._has_owner_access(request.user, expert_request.project.customer):
            raise exceptions.PermissionDenied()

    cancel_permissions = complete_permissions = [is_owner, is_valid_request]


class ExpertBidViewSet(core_views.ActionsViewSet):
    queryset = models.ExpertBid.objects.all()
    serializer_class = serializers.ExpertBidSerializer
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.ExpertBidFilter
    disabled_actions = ['update']

    def is_expert_manager(request, view, obj=None):
        if not is_expert_manager(request.user):
            raise exceptions.PermissionDenied()

    create_permissions = [is_expert_manager]

    def get_queryset(self):
        return super(ExpertBidViewSet, self).get_queryset() \
            .filtered_for_user(self.request.user).distinct()

    @decorators.detail_route(methods=['post'])
    @transaction.atomic()
    def accept(self, request, uuid=None):
        current_user = self.request.user
        expert_bid = self.get_object()

        expert_request = expert_bid.request
        expert_request.state = models.ExpertRequest.States.ACTIVE
        expert_request.save(update_fields=['state'])

        expert_contract = models.ExpertContract.objects.create(
            request=expert_request,
            team=expert_bid.team,
            price=expert_bid.price,
            description=expert_bid.description,
        )

        transaction.on_commit(lambda:
                              create_team_invitations(expert_bid.team, expert_request.project, current_user))

        transaction.on_commit(lambda: tasks.create_pdf_contract.delay(expert_contract.id))

        return response.Response({'status': _('Expert bid has been accepted.')}, status=status.HTTP_200_OK)

    def is_pending_request(request, view, obj=None):
        if obj and obj.request.state != models.ExpertRequest.States.PENDING:
            raise exceptions.ValidationError(_('Expert request should be in pending state.'))

    accept_permissions = [structure_permissions.is_owner, is_pending_request]

    def can_delete_bid(request, view, obj=None):
        if not obj:
            return
        if request.user.is_staff:
            return
        if not structure_permissions._has_owner_access(request.user, obj.team.customer):
            raise exceptions.PermissionDenied()

    destroy_permissions = [is_pending_request, can_delete_bid]


structure_views.ProjectCountersView.register_counter('experts', quotas.get_experts_count)
structure_views.CustomerCountersView.register_counter('experts', quotas.get_experts_count)


def experts_customers(user):
    connected_customers_query = structure_models.Customer.objects.all()
    connected_customers_query = connected_customers_query.filter(
        Q(permissions__user=user, permissions__is_active=True) |
        Q(projects__permissions__user=user, projects__permissions__is_active=True)
    ).distinct()

    expert_bids = models.ExpertBid.objects.filter(
        Q(request__customer__in=connected_customers_query) |
        Q(team__customer__in=connected_customers_query)
    ).distinct()

    relation_customers = structure_models.Customer.objects.filter(
        Q(projects__expertbid__in=expert_bids) |
        Q(expertrequest__bids__in=expert_bids)).distinct()

    return Q(customerpermission__customer__in=relation_customers, customerpermission__is_active=True) | \
        Q(projectpermission__project__customer__in=relation_customers, projectpermission__is_active=True)


structure_filters.UserFilterBackend.register_extra_query(experts_customers)



import logging
import sys

from celery.task import Task as CeleryTask
from django.utils import six

from waldur_core.core import utils
from waldur_core.core.tasks import BackendMethodTask, Task

from . import backend, log

logger = logging.getLogger(__name__)


class WaitForActionComplete(CeleryTask):
    max_retries = 300
    default_retry_delay = 5

    def run(self, action_id, serialized_droplet):
        droplet = utils.deserialize_instance(serialized_droplet)
        backend = droplet.get_backend()
        action = backend.manager.get_action(action_id)
        if action.status == 'completed':
            backend_droplet = backend.get_droplet(droplet.backend_id)
            droplet.ip_address = backend_droplet.ip_address
            droplet.save(update_fields=['ip_address'])
            return True
        else:
            self.retry()


class LogDropletResized(Task):

    def execute(self, droplet, serialized_size, *args, **kwargs):
        size = utils.deserialize_instance(serialized_size)
        logger.info('Successfully resized droplet %s', droplet.uuid.hex)
        log.event_logger.droplet_resize.info(
            'Droplet {droplet_name} has been resized.',
            event_type='droplet_resize_succeeded',
            event_context={'droplet': droplet, 'size': size}
        )


class SafeBackendMethodTask(BackendMethodTask):
    """
    Open alert if token scope is read-only.
    Close alert if token scope if read-write.
    It should be applied to droplet provisioning tasks.
    """

    def execute(self, droplet, *args, **kwargs):
        try:
            result = super(SafeBackendMethodTask, self).execute(droplet, *args, **kwargs)
        except backend.TokenScopeError:
            droplet.service_project_link.service.raise_readonly_token_alert()
            six.reraise(*sys.exc_info())
        else:
            droplet.service_project_link.service.close_readonly_token_alert()
            return result



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import taggit.managers
import django_fsm
import waldur_core.core.validators


def migrate_floatingip_status(apps, schema_editor):
    FloatingIP = apps.get_model('openstack', 'FloatingIP')

    for floating_ip in FloatingIP.objects.all():
        floating_ip.runtime_state = floating_ip.status or 'DOWN'
        floating_ip.save()


def migrate_floatingip_name(apps, schema_editor):
    FloatingIP = apps.get_model('openstack', 'FloatingIP')

    for floating_ip in FloatingIP.objects.all():
        floating_ip.name = '' if floating_ip.address is None else str(floating_ip.address)
        floating_ip.save()


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('openstack', '0025_delete_tenant_models'),
    ]

    operations = [
        migrations.AddField(
            model_name='floatingip',
            name='created',
            field=model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='description',
            field=models.CharField(max_length=500, verbose_name='description', blank=True),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='error_message',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='modified',
            field=model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='name',
            field=models.CharField(default='', max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name]),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='floatingip',
            name='runtime_state',
            field=models.CharField(max_length=150, verbose_name='runtime state', blank=True),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='start_time',
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='state',
            field=django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')]),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='tags',
            field=taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags'),
        ),
        migrations.AlterField(
            model_name='floatingip',
            name='address',
            field=models.GenericIPAddressField(null=True, protocol='IPv4', blank=True),
        ),
        migrations.AlterField(
            model_name='floatingip',
            name='backend_id',
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.RunPython(migrate_floatingip_status),
        migrations.RunPython(migrate_floatingip_name),
    ]



from django.utils import timezone

from waldur_core.core import utils as core_utils

MAPPING = {
    'cpu_usage': 'nc_cpu_usage',
    'gpu_usage': 'nc_gpu_usage',
    'ram_usage': 'nc_ram_usage',
}

FIELD_NAMES = MAPPING.keys()

QUOTA_NAMES = MAPPING.values()


def format_current_month():
    today = timezone.now()
    month_start = core_utils.month_start(today).strftime('%Y-%m-%d')
    month_end = core_utils.month_end(today).strftime('%Y-%m-%d')
    return month_start, month_end



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import decorators, permissions, response, status, viewsets

from waldur_core.structure import filters as structure_filters
from waldur_core.structure import views as structure_views
from waldur_core.structure import permissions as structure_permissions

from . import executors, filters, models, serializers


class SlurmServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.SlurmService.objects.all()
    serializer_class = serializers.ServiceSerializer


class SlurmServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.SlurmServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer
    filter_class = filters.SlurmServiceProjectLinkFilter


class AllocationViewSet(structure_views.BaseResourceViewSet):
    queryset = models.Allocation.objects.all()
    serializer_class = serializers.AllocationSerializer
    filter_class = filters.AllocationFilter

    create_executor = executors.AllocationCreateExecutor
    pull_executor = executors.AllocationPullExecutor

    destroy_permissions = [structure_permissions.is_staff]
    delete_executor = executors.AllocationDeleteExecutor

    partial_update_permissions = update_permissions = [structure_permissions.is_owner]
    update_executor = executors.AllocationUpdateExecutor

    @decorators.detail_route(methods=['post'])
    def cancel(self, request, uuid=None):
        allocation = self.get_object()
        allocation.get_backend().cancel_allocation(allocation)
        return response.Response(status=status.HTTP_200_OK)

    cancel_permissions = [structure_permissions.is_owner]


class AllocationUsageViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.AllocationUsage.objects.all()
    serializer_class = serializers.AllocationUsageSerializer
    lookup_field = 'uuid'
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.AllocationUsageFilter


def get_project_allocation_count(project):
    return project.quotas.get(name='nc_allocation_count').usage


structure_views.ProjectCountersView.register_counter('slurm', get_project_allocation_count)



from waldur_core.logging.loggers import AlertLogger, alert_logger, EventLogger, event_logger


class DigitalOceanAlertLogger(AlertLogger):
    settings = 'structure.ServiceSettings'

    class Meta:
        alert_types = ['token_is_read_only']


class DropletResizeEventLogger(EventLogger):
    droplet = 'waldur_digitalocean.Droplet'
    size = 'waldur_digitalocean.Size'

    class Meta:
        event_types = ('droplet_resize_scheduled',
                       'droplet_resize_succeeded')


alert_logger.register('digital_ocean', DigitalOceanAlertLogger)
event_logger.register('droplet_resize', DropletResizeEventLogger)



from django.apps import AppConfig
from django.db.models import signals


class MarketplacePackageConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_packages'
    verbose_name = 'Marketplace VPC packages'

    def ready(self):
        from waldur_mastermind.packages import models as package_models
        from waldur_mastermind.marketplace.plugins import manager

        from . import handlers, processor

        signals.post_save.connect(
            handlers.create_offering_for_package_template,
            sender=package_models.PackageTemplate,
            dispatch_uid='waldur_mastermind.marketpace_packages.'
                         'create_offering_for_package_template',
        )

        signals.post_save.connect(
            handlers.sync_offering_attribute_with_template_component,
            sender=package_models.PackageComponent,
            dispatch_uid='waldur_mastermind.marketpace_packages.'
                         'sync_offering_attribute_with_template_component',
        )

        manager.register(package_models.PackageTemplate, processor.process_order_item)



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals


class CostTrackingConfig(AppConfig):
    name = 'waldur_core.cost_tracking'
    verbose_name = 'Cost tracking'

    def ready(self):
        from waldur_core.cost_tracking import handlers
        from waldur_core.quotas import models as quotas_models
        from waldur_core.structure import models as structure_models

        PriceEstimate = self.get_model('PriceEstimate')

        for index, model in enumerate(PriceEstimate.get_estimated_models()):
            signals.pre_delete.connect(
                handlers.scope_deletion,
                sender=model,
                dispatch_uid='waldur_core.cost_tracking.handlers.scope_deletion_%s_%s' % (model.__name__, index),
            )

        for index, model in enumerate(structure_models.ResourceMixin.get_all_models()):
            signals.post_save.connect(
                handlers.resource_update,
                sender=model,
                dispatch_uid='waldur_core.cost_tracking.resource_update_%s_%s' % (model.__name__, index),
            )

        signals.post_save.connect(
            handlers.resource_quota_update,
            sender=quotas_models.Quota,
            dispatch_uid='waldur_core.cost_tracking.handlers.resource_quota_update',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-04-18 14:11
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0047_privateservicesettings_sharedservicesettings'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='balancehistory',
            name='customer',
        ),
        migrations.RemoveField(
            model_name='customer',
            name='balance',
        ),
        migrations.DeleteModel(
            name='BalanceHistory',
        ),
    ]



from . import tasks, utils
from .log import event_logger


def schedule_sync(*args, **kwargs):
    tasks.schedule_sync()


def schedule_sync_on_quota_change(sender, instance, created=False, **kwargs):
    if instance.name != utils.QUOTA_NAME:
        return
    if created and instance.limit == -1:
        return
    tasks.schedule_sync()


def log_profile_event(sender, instance, created=False, **kwargs):
    profile = instance

    if created:
        event_logger.freeipa.info(
            '{username} FreeIPA profile has been created.',
            event_type='freeipa_profile_created',
            event_context={
                'user': profile.user,
                'username': profile.username,
            }
        )

    elif profile.tracker.has_changed('is_active') and profile.tracker.previous('is_active'):
        event_logger.freeipa.info(
            '{username} FreeIPA profile has been disabled.',
            event_type='freeipa_profile_disabled',
            event_context={
                'user': profile.user,
                'username': profile.username,
            }
        )

    elif profile.tracker.has_changed('is_active') and not profile.tracker.previous('is_active'):
        event_logger.freeipa.info(
            '{username} FreeIPA profile has been enabled.',
            event_type='freeipa_profile_enabled',
            event_context={
                'user': profile.user,
                'username': profile.username,
            }
        )


def log_profile_deleted(sender, instance, **kwargs):
    profile = instance
    event_logger.freeipa.info(
        '{username} FreeIPA profile has been deleted.',
        event_type='freeipa_profile_deleted',
        event_context={
            'user': profile.user,
            'username': profile.username,
        }
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-16 09:48
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import taggit.managers
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('structure', '0052_customer_subnets'),
    ]

    operations = [
        migrations.CreateModel(
            name='Host',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=64, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('visible_name', models.CharField(max_length=64, verbose_name='visible name')),
                ('interface_parameters', waldur_core.core.fields.JSONField(blank=True)),
                ('host_group_name', models.CharField(blank=True, max_length=64, verbose_name='host group name')),
                ('error', models.CharField(blank=True, help_text='Error text if Zabbix agent is unavailable.', max_length=500)),
                ('status', models.CharField(choices=[('0', 'monitored'), ('1', 'unmonitored')], default='0', max_length=30)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('backend_id', models.CharField(max_length=64)),
                ('value_type', models.IntegerField(choices=[(0, 'Numeric (float)'), (1, 'Character'), (2, 'Log'), (3, 'Numeric (unsigned)'), (4, 'Text')])),
                ('units', models.CharField(max_length=255)),
                ('history', models.IntegerField()),
                ('delay', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='ITService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('is_main', models.BooleanField(default=True, help_text='Main IT service SLA will be added to hosts resource as monitoring item.')),
                ('algorithm', models.PositiveSmallIntegerField(choices=[(0, 'do not calculate'), (1, 'problem, if at least one child has a problem'), (2, 'problem, if all children have problems')], default=0)),
                ('sort_order', models.PositiveSmallIntegerField(default=1)),
                ('agreed_sla', models.DecimalField(blank=True, decimal_places=4, max_digits=6, null=True)),
                ('backend_trigger_id', models.CharField(blank=True, max_length=64, null=True)),
                ('host', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='itservices', to='waldur_zabbix.Host')),
            ],
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SlaHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('period', models.CharField(max_length=10)),
                ('value', models.DecimalField(blank=True, decimal_places=4, max_digits=11, null=True)),
                ('itservice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_zabbix.ITService')),
            ],
            options={
                'verbose_name': 'SLA history',
                'verbose_name_plural': 'SLA histories',
            },
        ),
        migrations.CreateModel(
            name='SlaHistoryEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.IntegerField()),
                ('state', models.CharField(choices=[('U', 'DOWN'), ('D', 'UP')], max_length=1)),
                ('history', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='events', to='waldur_zabbix.SlaHistory')),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('parents', models.ManyToManyField(related_name='children', to='waldur_zabbix.Template')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Trigger',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='triggers', to='waldur_zabbix.Template')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('alias', models.CharField(max_length=150)),
                ('surname', models.CharField(max_length=150)),
                ('type', models.CharField(choices=[('1', 'default'), ('2', 'admin'), ('3', 'superadmin')], default='1', max_length=30)),
                ('password', models.CharField(blank=True, max_length=150)),
                ('phone', models.CharField(blank=True, max_length=30)),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='UserGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ZabbixService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ZabbixServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_zabbix.ZabbixService')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='zabbixservice',
            name='projects',
            field=models.ManyToManyField(related_name='zabbix_services', through='waldur_zabbix.ZabbixServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='zabbixservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='user',
            name='groups',
            field=models.ManyToManyField(related_name='users', to='waldur_zabbix.UserGroup'),
        ),
        migrations.AddField(
            model_name='user',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='itservice',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='itservices', to='waldur_zabbix.ZabbixServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='itservice',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='itservice',
            name='trigger',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='waldur_zabbix.Trigger'),
        ),
        migrations.AddField(
            model_name='item',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='waldur_zabbix.Template'),
        ),
        migrations.AddField(
            model_name='host',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='hosts', to='waldur_zabbix.ZabbixServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='host',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='host',
            name='templates',
            field=models.ManyToManyField(related_name='hosts', to='waldur_zabbix.Template'),
        ),
        migrations.AlterUniqueTogether(
            name='zabbixserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='zabbixservice',
            unique_together=set([('customer', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='usergroup',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='user',
            unique_together=set([('alias', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='trigger',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='template',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='slahistory',
            unique_together=set([('itservice', 'period')]),
        ),
        migrations.AlterUniqueTogether(
            name='itservice',
            unique_together=set([('host', 'is_main')]),
        ),
    ]



from __future__ import unicode_literals

import copy
import logging
import os

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from django.template import Context, Template
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers, exceptions

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import models as structure_models, SupportedServices, serializers as structure_serializers
from waldur_jira.serializers import WebHookReceiverSerializer as JiraWebHookReceiverSerializer
from waldur_mastermind.common.mixins import UnitPriceMixin
from waldur_mastermind.support.backend.atlassian import ServiceDeskBackend

from . import models

logger = logging.getLogger(__name__)
User = get_user_model()


def render_issue_template(config_name, issue):
    issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})
    if not issue_settings:
        return ''

    raw = issue_settings[config_name]
    template = Template(raw)
    return template.render(Context({'issue': issue}))


class IssueSerializer(core_serializers.AugmentedSerializerMixin,
                      serializers.HyperlinkedModelSerializer):
    resource = core_serializers.GenericRelatedField(
        related_models=structure_models.ResourceMixin.get_all_models(), required=False)
    caller = serializers.HyperlinkedRelatedField(
        view_name='user-detail',
        lookup_field='uuid',
        queryset=User.objects.all(),
        required=False,
        allow_null=True,
    )
    reporter = serializers.HyperlinkedRelatedField(
        view_name='support-user-detail',
        lookup_field='uuid',
        read_only=True
    )
    assignee = serializers.HyperlinkedRelatedField(
        view_name='support-user-detail',
        lookup_field='uuid',
        queryset=models.SupportUser.objects.all(),
        required=False,
        allow_null=True,
    )
    resource_type = serializers.SerializerMethodField()
    resource_name = serializers.ReadOnlyField(source='resource.name')
    type = serializers.ChoiceField(
        choices=[(t, t) for t in settings.WALDUR_SUPPORT['ISSUE']['types']],
        initial=settings.WALDUR_SUPPORT['ISSUE']['types'][0],
        default=settings.WALDUR_SUPPORT['ISSUE']['types'][0])
    is_reported_manually = serializers.BooleanField(
        initial=False, default=False, write_only=True,
        help_text=_('Set true if issue is created by regular user via portal.'))

    class Meta(object):
        model = models.Issue
        fields = (
            'url', 'uuid', 'type', 'key', 'backend_id', 'link',
            'summary', 'description', 'status', 'resolution', 'priority',
            'caller', 'caller_uuid', 'caller_full_name',
            'reporter', 'reporter_uuid', 'reporter_name',
            'assignee', 'assignee_uuid', 'assignee_name',
            'customer', 'customer_uuid', 'customer_name',
            'project', 'project_uuid', 'project_name',
            'resource', 'resource_type', 'resource_name',
            'created', 'modified', 'is_reported_manually',
            'first_response_sla',
        )
        read_only_fields = ('key', 'status', 'resolution', 'backend_id', 'link', 'priority', 'first_response_sla')
        protected_fields = ('customer', 'project', 'resource', 'type', 'caller')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            customer={'lookup_field': 'uuid', 'view_name': 'customer-detail'},
            project={'lookup_field': 'uuid', 'view_name': 'project-detail'},
        )
        related_paths = dict(
            caller=('uuid', 'full_name',),
            reporter=('uuid', 'name',),
            assignee=('uuid', 'name',),
            customer=('uuid', 'name',),
            project=('uuid', 'name',),
        )

    def get_fields(self):
        fields = super(IssueSerializer, self).get_fields()

        if 'view' not in self.context:  # On docs generation context does not contain "view".
            return fields

        user = self.context['view'].request.user
        if not user.is_staff and not user.is_support:
            del fields['link']

        return fields

    def get_resource_type(self, obj):
        if obj.resource:
            return SupportedServices.get_name_for_model(obj.resource_content_type.model_class())

    def validate(self, attrs):
        if self.instance is not None:
            return attrs
        if attrs.pop('is_reported_manually'):
            attrs['caller'] = self.context['request'].user
            if attrs.get('assignee'):
                raise serializers.ValidationError(
                    {'assignee': _('Assignee cannot be defined if issue is reported manually.')})
        else:
            if not attrs.get('caller'):
                raise serializers.ValidationError({'caller': _('This field is required.')})
            reporter = models.SupportUser.objects.filter(
                user=self.context['request'].user, is_active=True).first()
            if not reporter:
                raise serializers.ValidationError(
                    _('You cannot report issues because your help desk account is not connected to profile.'))
            attrs['reporter'] = reporter
        return attrs

    def validate_customer(self, customer):
        """ User has to be customer owner, staff or global support """
        if not customer:
            return customer
        user = self.context['request'].user
        if (not customer or
                user.is_staff or
                user.is_support or
                customer.has_user(user, structure_models.CustomerRole.OWNER)):
            return customer
        raise serializers.ValidationError(_('Only customer owner, staff or support can report customer issues.'))

    def validate_project(self, project):
        if not project:
            return project
        user = self.context['request'].user
        if (not project or
                user.is_staff or
                user.is_support or
                project.customer.has_user(user, structure_models.CustomerRole.OWNER) or
                project.has_user(user, structure_models.ProjectRole.MANAGER) or
                project.has_user(user, structure_models.ProjectRole.ADMINISTRATOR)):
            return project
        raise serializers.ValidationError(
            _('Only customer owner, project manager, project admin, staff or support can report such issue.'))

    def validate_resource(self, resource):
        if resource:
            self.validate_project(resource.service_project_link.project)
        return resource

    @transaction.atomic()
    def create(self, validated_data):
        resource = validated_data.get('resource')
        if resource:
            validated_data['project'] = resource.service_project_link.project
        project = validated_data.get('project')
        if project:
            validated_data['customer'] = project.customer

        validated_data['description'] = render_issue_template('description', validated_data)
        validated_data['summary'] = render_issue_template('summary', validated_data)
        return super(IssueSerializer, self).create(validated_data)

    def _render_template(self, config_name, issue):
        raw = self.issue_settings[config_name]
        template = Template(raw)
        return template.render(Context({'issue': issue}))


class CommentSerializer(core_serializers.AugmentedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):
    # should be initialized with issue in context on creation
    author_user = serializers.HyperlinkedRelatedField(
        source='author.user',
        view_name='user-detail',
        lookup_field='uuid',
        read_only=True,
    )

    author_uuid = serializers.ReadOnlyField(source='author.user.uuid')

    class Meta(object):
        model = models.Comment
        fields = ('url', 'uuid', 'issue', 'issue_key', 'description', 'is_public',
                  'author_name', 'author_uuid', 'author_user', 'backend_id', 'created')
        read_only_fields = ('issue', 'backend_id',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
        )
        related_paths = dict(
            author=('name',),
            issue=('key',),
        )

    @transaction.atomic()
    def create(self, validated_data):
        author_user = self.context['request'].user
        validated_data['author'], _ = models.SupportUser.objects.get_or_create_from_user(author_user)
        validated_data['issue'] = self.context['view'].get_object()
        return super(CommentSerializer, self).create(validated_data)


class SupportUserSerializer(core_serializers.AugmentedSerializerMixin,
                            serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.SupportUser
        fields = ('url', 'uuid', 'name', 'backend_id', 'user')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            user={'lookup_field': 'uuid', 'view_name': 'user-detail'}
        )


class WebHookReceiverSerializer(JiraWebHookReceiverSerializer):
    def get_project(self, project_key):
        class Project:
            def get_backend(self):
                return ServiceDeskBackend()

        return Project()

    def get_issue(self, project, key, create):
        issue = None

        try:
            issue = models.Issue.objects.get(backend_id=key)
        except models.Issue.DoesNotExist:
            if not create:
                raise serializers.ValidationError('Issue with id %s does not exist.' % key)

        return issue

    def get_comment(self, issue, key, create):
        comment = None

        try:
            comment = models.Comment.objects.get(issue=issue, backend_id=key)
        except models.Comment.DoesNotExist:
            if not create:
                raise serializers.ValidationError('Comment with id %s does not exist.' % key)

        return comment


WebHookReceiverSerializer.remove_event(['jira:issue_created'])


class OfferingSerializer(structure_serializers.PermissionFieldFilteringMixin,
                         core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    type = serializers.ChoiceField(choices=settings.WALDUR_SUPPORT['OFFERINGS'].keys())
    state = serializers.ReadOnlyField(source='get_state_display')
    report = serializers.JSONField(required=False)

    class Meta(object):
        model = models.Offering
        fields = ('url', 'uuid', 'name', 'project', 'type', 'state', 'type_label', 'unit_price',
                  'unit', 'created', 'modified', 'issue', 'issue_name', 'issue_link',
                  'issue_key', 'issue_description', 'issue_uuid', 'issue_status',
                  'project_name', 'project_uuid', 'product_code', 'article_code', 'report')
        read_only_fields = ('type_label', 'issue', 'unit_price', 'unit', 'state', 'product_code', 'article_code')
        protected_fields = ('project', 'type')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
            project={'lookup_field': 'uuid', 'view_name': 'project-detail'},
            unit_price={'decimal_places': 2},
        )
        related_paths = dict(
            issue=('uuid', 'name', 'status', 'key', 'description', 'link'),
            project=('uuid', 'name',),
        )

    def validate_report(self, report):
        if not isinstance(report, list):
            raise serializers.ValidationError('Report should be an object.')

        if len(report) == 0:
            raise serializers.ValidationError('Report object should contain at least one section.')

        for section in report:
            if not isinstance(section, dict):
                raise serializers.ValidationError('Report section should be an object.')

            if not section.get('header'):
                raise serializers.ValidationError('Report section should contain header.')

            if not section.get('body'):
                raise serializers.ValidationError('Report section should contain body.')

        return report

    def get_filtered_field_names(self):
        return ('project',)


class ConfigurableSerializerMixin(object):
    """
    Serializer is built on top WALDUR_SUPPORT['OFFERINGS'] configuration.

    Each configured field get's converted to serializer field according to field type in the configuration.

    Field types:
        'integer' - corresponds to 'serializers.IntegerField'
        'string' - is a default field type even if it is not defined explicitly in configuration.
                   Corresponds to 'serializers.CharField(max_length=255)'

    Default values:
        if 'default' key is present in option field configuration it is going to be used in serializer unless
        the value itself has been provided in the create request.
    """

    def _get_offerings_configuration(self):
        return copy.deepcopy(settings.WALDUR_SUPPORT['OFFERINGS'])

    def _get_configuration(self, type):
        return self._get_offerings_configuration().get(type)

    def get_fields(self):
        result = super(ConfigurableSerializerMixin, self).get_fields()
        if hasattr(self, 'initial_data') and not hasattr(self, '_errors'):
            type = self.initial_data['type']
            configuration = self._get_configuration(type)
            for attr_name in configuration['order']:
                attr_options = configuration['options'].get(attr_name, {})
                result[attr_name] = self._get_field_instance(attr_options)

        # choices have to be added dynamically so that unit tests can mock offering configuration.
        # otherwise it is always going to be a default set up.
        result['type'] = serializers.ChoiceField(allow_blank=False, choices=self._get_offerings_configuration().keys())

        return result

    def _validate_type(self, type):
        offering_configuration = self._get_configuration(type)
        if offering_configuration is None:
            raise serializers.ValidationError({
                'type': _('Type configuration could not be found.')
            })

    def validate_empty_values(self, data):
        if 'type' not in data or ('type' in data and data['type'] is None):
            raise serializers.ValidationError({
                'type': _('This field is required.')
            })
        else:
            self._validate_type(data['type'])

        return super(ConfigurableSerializerMixin, self).validate_empty_values(data)

    def _get_field_instance(self, attr_options):
        field_type = attr_options.get('type', '').lower()

        if field_type == 'string':
            field = serializers.CharField(max_length=255, write_only=True)
        elif field_type == 'integer':
            field = serializers.IntegerField(write_only=True)
        else:
            field = serializers.CharField(write_only=True)

        default_value = attr_options.get('default')
        if default_value:
            field.default = default_value

        field.required = attr_options.get('required', False)
        field.label = attr_options.get('label')
        field.help_text = attr_options.get('help_text')

        return field

    def _form_description(self, configuration, validated_data):
        result = []

        for key in configuration['order']:
            if key not in validated_data:
                continue

            label = configuration['options'].get(key, {})
            label_value = label.get('label', key)
            result.append('%s: \'%s\'' % (label_value, validated_data[key]))

        if 'description' in validated_data:
            result.append('\n %s' % validated_data['description'])

        return '\n'.join(result)

    def _get_extra(self, configuration, validated_data):
        result = {
            key: validated_data[key]
            for key in configuration['order']
            if key in validated_data
        }
        if 'description' in validated_data:
            result['description'] = validated_data['description']
        return result


class OfferingCreateSerializer(ConfigurableSerializerMixin, OfferingSerializer):
    description = serializers.CharField(required=False, help_text=_('Description to add to the issue.'))

    class Meta(OfferingSerializer.Meta):
        fields = OfferingSerializer.Meta.fields + ('description',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
            project={'lookup_field': 'uuid',
                     'view_name': 'project-detail',
                     'required': True,
                     'allow_empty': False,
                     'allow_null': False},
        )

    def create(self, validated_data):
        """
        Each offering corresponds to the single issue which has next values:
            'project' - a hyperlinked field which must be provided with every request;
            'customer' - customer is extracted from the provided project;
            'caller' - a user who sent a request is considered to be a 'caller' of the issue;
            'summary' - has a format of 'Request for "OFFERING[name][label]' or 'Request for "Support" if empty;
            'description' - combined list of all other fields provided with the request;
        """
        project = validated_data['project']
        type = validated_data['type']
        offering_configuration = self._get_configuration(type)
        type_label = offering_configuration.get('label', type)
        issue_details = dict(
            caller=self.context['request'].user,
            project=project,
            customer=project.customer,
            type=settings.WALDUR_SUPPORT['DEFAULT_OFFERING_ISSUE_TYPE'],
            summary='Request for \'%s\'' % type_label,
            description=self._form_description(offering_configuration, validated_data))
        issue_details['summary'] = render_issue_template('summary', issue_details)
        issue_details['description'] = render_issue_template('description', issue_details)
        issue = models.Issue.objects.create(**issue_details)

        offering = models.Offering.objects.create(
            issue=issue,
            project=issue.project,
            name=validated_data.get('name'),
            type=type,
            product_code=offering_configuration.get('product_code', ''),
            article_code=offering_configuration.get('article_code', ''),
            unit_price=offering_configuration.get('price', 0),
            unit=offering_configuration.get('unit', UnitPriceMixin.Units.PER_MONTH),
        )

        return offering


class OfferingCompleteSerializer(serializers.Serializer):
    unit_price = serializers.DecimalField(max_digits=13, decimal_places=7)
    unit = serializers.ChoiceField(choices=models.Offering.Units.CHOICES, default=models.Offering.Units.PER_DAY)

    def update(self, instance, validated_data):
        instance.unit_price = validated_data['unit_price']
        instance.unit = validated_data['unit']
        instance.state = models.Offering.States.OK
        instance.save(update_fields=['state', 'unit_price', 'unit'])
        return instance


class AttachmentSerializer(core_serializers.AugmentedSerializerMixin,
                           serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.Attachment
        fields = ('url', 'uuid', 'issue', 'issue_key', 'created', 'file',
                  'mime_type', 'file_size', 'thumbnail', 'backend_id', )
        read_only_fields = ('backend_id',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
        )
        related_paths = dict(
            issue=('key',),
        )

    def validate(self, attrs):
        filename, file_extension = os.path.splitext(attrs['file'].name)
        if file_extension in settings.WALDUR_SUPPORT['EXCLUDED_ATTACHMENT_TYPES']:
            raise serializers.ValidationError(_('Invalid file extension'))

        user = self.context['request'].user
        issue = attrs['issue']
        if user.is_staff or \
                (issue.customer and issue.customer.has_user(user, structure_models.CustomerRole.OWNER)) or \
                issue.caller == user:
            return attrs

        raise exceptions.PermissionDenied()


class TemplateAttachmentSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.TemplateAttachment
        fields = ('name', 'file')


class TemplateSerializer(serializers.HyperlinkedModelSerializer):
    attachments = TemplateAttachmentSerializer(many=True)

    class Meta(object):
        model = models.Template
        fields = ('url', 'uuid',
                  'name', 'native_name',
                  'description', 'native_description',
                  'issue_type', 'attachments')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-template-detail'},
        )

    def get_fields(self):
        fields = super(TemplateSerializer, self).get_fields()
        if not settings.WALDUR_CORE['NATIVE_NAME_ENABLED']:
            del fields['native_name']
            del fields['native_description']
        return fields



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class MarketplacePackagesExtension(WaldurExtension):

    class Settings:
        WALDUR_MARKETPLACE_PACKAGES = {
            'CUSTOMER_ID': None,
            'CATEGORY_ID': None,
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_packages'

    @staticmethod
    def is_assembly():
        return True



def update_registration_method(sender, instance, **kwargs):
    user = instance
    if user.registration_method != 'SAML2':
        user.registration_method = 'SAML2'
        return True
    else:
        return False
